# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import List, Optional

from attr import dataclass
import pytest

from ..primitive import JSON
from .serializable_attrs import Serializable, SerializableAttrs, SerializerError, field


def test_simple_class():
    @dataclass
    class Foo(SerializableAttrs):
        hello: int
        world: str

    serialized = {"hello": 5, "world": "hi"}
    deserialized = Foo.deserialize(serialized)
    assert deserialized == Foo(5, "hi")
    assert deserialized.serialize() == serialized

    with pytest.raises(SerializerError):
        Foo.deserialize({"world": "hi"})


def test_default():
    @dataclass
    class Default(SerializableAttrs):
        no_default: int
        defaultful_value: int = 5

    d1 = Default.deserialize({"no_default": 3})
    assert d1.no_default == 3
    assert d1.defaultful_value == 5
    d2 = Default.deserialize({"no_default": 4, "defaultful_value": 6})
    assert d2.no_default == 4
    assert d2.defaultful_value == 6


def test_factory():
    @dataclass
    class Factory(SerializableAttrs):
        manufactured_value: List[str] = field(factory=lambda: ["hi"])

    assert Factory.deserialize({}).manufactured_value == ["hi"]
    assert Factory.deserialize({"manufactured_value": None}).manufactured_value == ["hi"]
    factory1 = Factory.deserialize({})
    factory2 = Factory.deserialize({})
    assert factory1.manufactured_value is not factory2.manufactured_value
    assert Factory.deserialize({"manufactured_value": ["bye"]}).manufactured_value == ["bye"]


def test_hidden():
    @dataclass
    class HiddenField(SerializableAttrs):
        visible: str
        hidden: int = field(hidden=True, default=5)

    deserialized_hidden = HiddenField.deserialize({"visible": "yay", "hidden": 4})
    assert deserialized_hidden.hidden == 5
    assert deserialized_hidden.unrecognized_["hidden"] == 4
    assert HiddenField("hmm", 5).serialize() == {"visible": "hmm"}


def test_ignore_errors():
    @dataclass
    class Something(SerializableAttrs):
        required: bool

    @dataclass
    class Wrapper(SerializableAttrs):
        something: Optional[Something] = field(ignore_errors=True)

    @dataclass
    class ErroringWrapper(SerializableAttrs):
        something: Optional[Something] = field(ignore_errors=False)

    assert Wrapper.deserialize({"something": {"required": True}}) == Wrapper(Something(True))
    assert Wrapper.deserialize({"something": {}}) == Wrapper(None)
    with pytest.raises(SerializerError):
        ErroringWrapper.deserialize({"something": 5})
    with pytest.raises(SerializerError):
        ErroringWrapper.deserialize({"something": {}})


def test_json_key_override():
    @dataclass
    class Meow(SerializableAttrs):
        meow: int = field(json="fi.mau.namespaced_meow")

    serialized = {"fi.mau.namespaced_meow": 123}
    deserialized = Meow.deserialize(serialized)
    assert deserialized == Meow(123)
    assert deserialized.serialize() == serialized


def test_omit_empty():
    @dataclass
    class OmitEmpty(SerializableAttrs):
        omitted: Optional[int] = field(omit_empty=True)

    @dataclass
    class DontOmitEmpty(SerializableAttrs):
        not_omitted: Optional[int] = field(omit_empty=False)

    assert OmitEmpty(None).serialize() == {}
    assert OmitEmpty(0).serialize() == {"omitted": 0}
    assert DontOmitEmpty(None).serialize() == {"not_omitted": None}
    assert DontOmitEmpty(0).serialize() == {"not_omitted": 0}


def test_omit_default():
    @dataclass
    class OmitDefault(SerializableAttrs):
        omitted: int = field(default=5, omit_default=True)

    @dataclass
    class DontOmitDefault(SerializableAttrs):
        not_omitted: int = 5

    assert OmitDefault().serialize() == {}
    assert OmitDefault(5).serialize() == {}
    assert OmitDefault(6).serialize() == {"omitted": 6}
    assert DontOmitDefault().serialize() == {"not_omitted": 5}
    assert DontOmitDefault(5).serialize() == {"not_omitted": 5}
    assert DontOmitDefault(6).serialize() == {"not_omitted": 6}


def test_flatten():
    from mautrix.types import ContentURI

    @dataclass
    class OpenGraphImage(SerializableAttrs):
        url: ContentURI = field(default=None, json="og:image")
        mimetype: str = field(default=None, json="og:image:type")
        height: int = field(default=None, json="og:image:width")
        width: int = field(default=None, json="og:image:height")
        size: int = field(default=None, json="matrix:image:size")

    @dataclass
    class OpenGraphVideo(SerializableAttrs):
        url: ContentURI = field(default=None, json="og:video")
        mimetype: str = field(default=None, json="og:video:type")
        height: int = field(default=None, json="og:video:width")
        width: int = field(default=None, json="og:video:height")
        size: int = field(default=None, json="matrix:video:size")

    @dataclass
    class OpenGraphAudio(SerializableAttrs):
        url: ContentURI = field(default=None, json="og:audio")
        mimetype: str = field(default=None, json="og:audio:type")

    @dataclass
    class MXOpenGraph(SerializableAttrs):
        title: str = field(default=None, json="og:title")
        description: str = field(default=None, json="og:description")
        image: OpenGraphImage = field(default=None, flatten=True)
        video: OpenGraphVideo = field(default=None, flatten=True)
        audio: OpenGraphAudio = field(default=None, flatten=True)

    example_com_preview = {
        "og:title": "Example Domain",
        "og:description": "Example Domain\n\nThis domain is for use in illustrative examples in "
        "documents. You may use this domain in literature without prior "
        "coordination or asking for permission.\n\nMore information...",
    }
    google_com_preview = {
        "og:title": "Google",
        "og:image": "mxc://maunium.net/2021-06-20_jkscuJXkHjvzNaUJ",
        "og:description": "Search\n\nImages\n\nMaps\n\nPlay\n\nYouTube\n\nNews\n\nGmail\n\n"
        "Drive\n\nMore\n\n\u00bb\n\nWeb History\n\n|\n\nSettings\n\n|\n\n"
        "Sign in\n\nAdvanced search\n\nGoogle offered in:\n\nDeutsch\n\n"
        "AdvertisingPrograms\n\nBusiness Solutions",
        "og:image:width": 128,
        "og:image:height": 128,
        "og:image:type": "image/png",
        "matrix:image:size": 3428,
    }

    example_com_deserialized = MXOpenGraph.deserialize(example_com_preview)
    assert example_com_deserialized.title == "Example Domain"
    assert example_com_deserialized.image is None
    assert example_com_deserialized.video is None
    assert example_com_deserialized.audio is None
    google_com_deserialized = MXOpenGraph.deserialize(google_com_preview)
    assert google_com_deserialized.title == "Google"
    assert google_com_deserialized.image is not None
    assert google_com_deserialized.image.url == "mxc://maunium.net/2021-06-20_jkscuJXkHjvzNaUJ"
    assert google_com_deserialized.image.width == 128
    assert google_com_deserialized.image.height == 128
    assert google_com_deserialized.image.mimetype == "image/png"
    assert google_com_deserialized.image.size == 3428
    assert google_com_deserialized.video is None
    assert google_com_deserialized.audio is None


def test_flatten_arbitrary_serializable():
    @dataclass
    class CustomSerializable(Serializable):
        is_hello: bool = True

        def serialize(self) -> JSON:
            if self.is_hello:
                return {"hello": "world"}
            return {}

        @classmethod
        def deserialize(cls, raw: JSON) -> "CustomSerializable":
            return CustomSerializable(is_hello=raw.get("hello") == "world")

    @dataclass
    class Thing(SerializableAttrs):
        custom: CustomSerializable = field(flatten=True)
        another: int = field(default=5)

    thing_1 = {
        "hello": "world",
        "another": 6,
    }
    thing_1_deserialized = Thing.deserialize(thing_1)
    assert thing_1_deserialized.custom.is_hello is True
    assert thing_1_deserialized.another == 6

    thing_2 = {
        "another": 4,
    }
    thing_2_deserialized = Thing.deserialize(thing_2)
    assert thing_2_deserialized.custom.is_hello is False
    assert thing_2_deserialized.another == 4

    assert Thing(custom=CustomSerializable(is_hello=True)).serialize() == {
        "hello": "world",
        "another": 5,
    }


def test_flatten_optional():
    @dataclass
    class OptionalThing(SerializableAttrs):
        key: str

        @classmethod
        def deserialize(cls, data: JSON) -> Optional["OptionalThing"]:
            if "key" in data:
                return super().deserialize(data)
            return None

    @dataclass
    class ThingWithOptional(SerializableAttrs):
        optional: OptionalThing = field(flatten=True)
        another_field: int = 2

    assert ThingWithOptional.deserialize({}).optional is None
    assert ThingWithOptional.deserialize({"key": "hi"}).optional.key == "hi"



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from .handler import SECTION_ADMIN, CommandEvent, command_handler


@command_handler(
    needs_auth=False,
    needs_puppeting=False,
    needs_admin=True,
    help_section=SECTION_ADMIN,
    help_text="Remove all users from the current portal room and forget the portal.",
)
async def delete_portal(evt: CommandEvent) -> None:
    if not evt.portal:
        await evt.reply("This is not a portal room")
        return
    await evt.portal.cleanup_and_delete()


@command_handler(
    needs_auth=False,
    needs_puppeting=False,
    help_section=SECTION_ADMIN,
    help_text="Remove puppets from the current portal room and forget the portal.",
)
async def unbridge(evt: CommandEvent) -> None:
    if not evt.portal:
        await evt.reply("This is not a portal room")
        return
    await evt.portal.unbridge()



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Any, Callable, Generic, Type, TypeVar
import re

from ...types import EventID, MatrixURI, RoomAlias, RoomID, UserID
from .formatted_string import EntityType, FormattedString
from .html_reader import HTMLNode, read_html
from .markdown_string import MarkdownString


class RecursionContext:
    preserve_whitespace: bool
    ul_depth: int
    _inited: bool

    def __init__(self, preserve_whitespace: bool = False, ul_depth: int = 0) -> None:
        self.preserve_whitespace = preserve_whitespace
        self.ul_depth = ul_depth
        self._inited = True

    def __setattr__(self, key: str, value: Any) -> None:
        if getattr(self, "_inited", False) is True:
            raise TypeError("'RecursionContext' object is immutable")
        super(RecursionContext, self).__setattr__(key, value)

    def enter_list(self) -> RecursionContext:
        return RecursionContext(
            preserve_whitespace=self.preserve_whitespace, ul_depth=self.ul_depth + 1
        )

    def enter_code_block(self) -> RecursionContext:
        return RecursionContext(preserve_whitespace=True, ul_depth=self.ul_depth)


T = TypeVar("T", bound=FormattedString)
spaces = re.compile(r"\s+")
space = " "


class MatrixParser(Generic[T]):
    block_tags: tuple[str, ...] = (
        "p",
        "pre",
        "blockquote",
        "ol",
        "ul",
        "li",
        "h1",
        "h2",
        "h3",
        "h4",
        "h5",
        "h6",
        "div",
        "hr",
        "table",
    )
    list_bullets: tuple[str, ...] = ("●", "○", "■", "‣")
    e: Type[EntityType] = EntityType
    fs: Type[T] = MarkdownString
    read_html: Callable[[str], HTMLNode] = staticmethod(read_html)
    ignore_less_relevant_links: bool = True
    exclude_plaintext_attrib: str = "data-mautrix-exclude-plaintext"

    def list_bullet(self, depth: int) -> str:
        return self.list_bullets[(depth - 1) % len(self.list_bullets)] + " "

    async def list_to_fstring(self, node: HTMLNode, ctx: RecursionContext) -> T:
        ordered: bool = node.tag == "ol"
        tagged_children: list[tuple[T, str]] = await self.node_to_tagged_fstrings(node, ctx)
        counter: int = 1
        indent_length: int = 0
        if ordered:
            try:
                counter = int(node.attrib.get("start", "1"))
            except ValueError:
                counter = 1

            longest_index = counter - 1 + len(tagged_children)
            indent_length = len(str(longest_index))
        indent: str = (indent_length + 2) * " "
        children: list[T] = []
        for child, tag in tagged_children:
            if tag != "li":
                continue

            if ordered:
                prefix = f"{counter}. "
                counter += 1
            else:
                prefix = self.list_bullet(ctx.ul_depth)
            child = child.prepend(prefix)
            parts = child.split("\n")
            parts = parts[:1] + [part.prepend(indent) for part in parts[1:]]
            child = self.fs.join(parts, "\n")
            children.append(child)
        return self.fs.join(children, "\n")

    async def blockquote_to_fstring(self, node: HTMLNode, ctx: RecursionContext) -> T:
        msg = await self.tag_aware_parse_node(node, ctx)
        return msg.format(self.e.BLOCKQUOTE)

    async def hr_to_fstring(self, node: HTMLNode, ctx: RecursionContext) -> T:
        return self.fs("---")

    async def header_to_fstring(self, node: HTMLNode, ctx: RecursionContext) -> T:
        children = await self.node_to_fstrings(node, ctx)
        length = int(node.tag[1])
        return self.fs.join(children, "").format(self.e.HEADER, size=length)

    async def basic_format_to_fstring(self, node: HTMLNode, ctx: RecursionContext) -> T:
        msg = await self.tag_aware_parse_node(node, ctx)
        if self.exclude_plaintext_attrib in node.attrib:
            return msg
        if node.tag in ("b", "strong"):
            msg = msg.format(self.e.BOLD)
        elif node.tag in ("i", "em"):
            msg = msg.format(self.e.ITALIC)
        elif node.tag in ("s", "strike", "del"):
            msg = msg.format(self.e.STRIKETHROUGH)
        elif node.tag in ("u", "ins"):
            msg = msg.format(self.e.UNDERLINE)
        return msg

    async def link_to_fstring(self, node: HTMLNode, ctx: RecursionContext) -> T:
        msg = await self.tag_aware_parse_node(node, ctx)
        href = node.attrib.get("href", "")
        if not href:
            return msg

        if href.startswith("mailto:"):
            return self.fs(href[len("mailto:") :]).format(self.e.EMAIL)

        matrix_uri = MatrixURI.try_parse(href)
        if matrix_uri:
            if matrix_uri.user_id:
                new_msg = await self.user_pill_to_fstring(msg, matrix_uri.user_id)
            elif matrix_uri.event_id:
                new_msg = await self.event_link_to_fstring(
                    msg, matrix_uri.room_id or matrix_uri.room_alias, matrix_uri.event_id
                )
            elif matrix_uri.room_alias:
                new_msg = await self.room_pill_to_fstring(msg, matrix_uri.room_alias)
            elif matrix_uri.room_id:
                new_msg = await self.room_id_link_to_fstring(msg, matrix_uri.room_id)
            else:
                new_msg = None
            if new_msg:
                return new_msg
        # Custom attribute to tell the parser that the link isn't relevant and
        # shouldn't be included in plaintext representation.
        if self.ignore_less_relevant_links and self.exclude_plaintext_attrib in node.attrib:
            return msg

        return await self.url_to_fstring(msg, href)

    async def url_to_fstring(self, msg: T, url: str) -> T | None:
        return msg.format(self.e.URL, url=url)

    async def user_pill_to_fstring(self, msg: T, user_id: UserID) -> T | None:
        return msg.format(self.e.USER_MENTION, user_id=user_id)

    async def room_pill_to_fstring(self, msg: T, room_alias: RoomAlias) -> T | None:
        return None

    async def room_id_link_to_fstring(self, msg: T, room_id: RoomID) -> T | None:
        return None

    async def event_link_to_fstring(
        self, msg: T, room: RoomID | RoomAlias, event_id: EventID
    ) -> T | None:
        return None

    async def img_to_fstring(self, node: HTMLNode, ctx: RecursionContext) -> T:
        return self.fs(node.attrib.get("alt") or node.attrib.get("title") or "")

    async def custom_node_to_fstring(self, node: HTMLNode, ctx: RecursionContext) -> T | None:
        return None

    async def color_to_fstring(self, msg: T, color: str) -> T:
        return msg.format(self.e.COLOR, color=color)

    async def spoiler_to_fstring(self, msg: T, reason: str) -> T:
        return msg.format(self.e.SPOILER, reason=reason)

    async def node_to_fstring(self, node: HTMLNode, ctx: RecursionContext) -> T:
        custom = await self.custom_node_to_fstring(node, ctx)
        if custom:
            return custom
        elif node.tag == "mx-reply":
            return self.fs("")
        elif node.tag == "blockquote":
            return await self.blockquote_to_fstring(node, ctx)
        elif node.tag == "hr":
            return await self.hr_to_fstring(node, ctx)
        elif node.tag == "ol":
            return await self.list_to_fstring(node, ctx)
        elif node.tag == "ul":
            return await self.list_to_fstring(node, ctx.enter_list())
        elif node.tag in ("h1", "h2", "h3", "h4", "h5", "h6"):
            return await self.header_to_fstring(node, ctx)
        elif node.tag == "br":
            return self.fs("\n")
        elif node.tag in ("b", "strong", "i", "em", "s", "del", "u", "ins"):
            return await self.basic_format_to_fstring(node, ctx)
        elif node.tag == "a":
            return await self.link_to_fstring(node, ctx)
        elif node.tag == "img":
            return await self.img_to_fstring(node, ctx)
        elif node.tag == "p":
            return (await self.tag_aware_parse_node(node, ctx)).append("\n")
        elif node.tag in ("font", "span"):
            msg = await self.tag_aware_parse_node(node, ctx)
            try:
                spoiler = node.attrib["data-mx-spoiler"]
            except KeyError:
                pass
            else:
                msg = await self.spoiler_to_fstring(msg, spoiler)

            try:
                color = node.attrib["color"]
            except KeyError:
                try:
                    color = node.attrib["data-mx-color"]
                except KeyError:
                    color = None
            if color:
                msg = await self.color_to_fstring(msg, color)
            return msg
        elif node.tag == "pre":
            lang = ""
            try:
                if node[0].tag == "code":
                    node = node[0]
                    lang = node.attrib["class"][len("language-") :]
            except (IndexError, KeyError):
                pass
            return (await self.parse_node(node, ctx.enter_code_block())).format(
                self.e.PREFORMATTED, language=lang
            )
        elif node.tag == "code":
            return (await self.parse_node(node, ctx.enter_code_block())).format(self.e.INLINE_CODE)
        return await self.tag_aware_parse_node(node, ctx)

    async def text_to_fstring(
        self, text: str, ctx: RecursionContext, strip_leading_whitespace: bool = False
    ) -> T:
        if not ctx.preserve_whitespace:
            text = spaces.sub(space, text.lstrip() if strip_leading_whitespace else text)
        return self.fs(text)

    async def node_to_tagged_fstrings(
        self, node: HTMLNode, ctx: RecursionContext
    ) -> list[tuple[T, str]]:
        output = []

        if node.text:
            output.append((await self.text_to_fstring(node.text, ctx), "text"))
        for child in node:
            output.append((await self.node_to_fstring(child, ctx), child.tag))
            if child.tail:
                # For text following a block tag, the leading whitespace is meaningless (there'll
                # be a newline added later), but for other tags it can be interpreted as a space.
                text = await self.text_to_fstring(
                    child.tail, ctx, strip_leading_whitespace=child.tag in self.block_tags
                )
                output.append((text, "text"))
        return output

    async def node_to_fstrings(self, node: HTMLNode, ctx: RecursionContext) -> list[T]:
        return [msg for (msg, tag) in await self.node_to_tagged_fstrings(node, ctx)]

    async def tag_aware_parse_node(self, node: HTMLNode, ctx: RecursionContext) -> T:
        msgs = await self.node_to_tagged_fstrings(node, ctx)
        output = self.fs()
        prev_was_block = False
        for msg, tag in msgs:
            if tag in self.block_tags:
                msg = msg.append("\n")
                if not prev_was_block:
                    msg = msg.prepend("\n")
                prev_was_block = True
            output = output.append(msg)
        return output.trim()

    async def parse_node(self, node: HTMLNode, ctx: RecursionContext) -> T:
        return self.fs.join(await self.node_to_fstrings(node, ctx))

    async def parse(self, data: str) -> T:
        msg = await self.node_to_fstring(
            self.read_html(f"<body>{data}</body>"), RecursionContext()
        )
        return msg



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from collections import defaultdict
import asyncio
import logging
import sys
import time

from mautrix import __optional_imports__
from mautrix.appservice import DOUBLE_PUPPET_SOURCE_KEY, AppService
from mautrix.errors import (
    DecryptionError,
    IntentError,
    MatrixError,
    MExclusive,
    MForbidden,
    MUnknownToken,
    SessionNotFound,
)
from mautrix.types import (
    BaseRoomEvent,
    BeeperMessageStatusEventContent,
    EncryptedEvent,
    Event,
    EventID,
    EventType,
    MediaRepoConfig,
    Membership,
    MemberStateEventContent,
    MessageEvent,
    MessageEventContent,
    MessageStatus,
    MessageStatusReason,
    MessageType,
    PresenceEvent,
    ReactionEvent,
    ReceiptEvent,
    ReceiptType,
    RedactionEvent,
    RelatesTo,
    RelationType,
    RoomID,
    RoomType,
    SingleReceiptEventContent,
    SpecVersions,
    StateEvent,
    StateUnsigned,
    TextMessageEventContent,
    TrustState,
    TypingEvent,
    UserID,
    Version,
    VersionsResponse,
)
from mautrix.util import background_task, markdown
from mautrix.util.logging import TraceLogger
from mautrix.util.message_send_checkpoint import (
    CHECKPOINT_TYPES,
    MessageSendCheckpoint,
    MessageSendCheckpointReportedBy,
    MessageSendCheckpointStatus,
    MessageSendCheckpointStep,
)
from mautrix.util.opt_prometheus import Histogram

from .. import bridge as br
from . import commands as cmd

encryption_import_error = None
media_encrypt_import_error = None

try:
    from .e2ee import EncryptionManager
except ImportError as e:
    if __optional_imports__:
        raise
    encryption_import_error = e
    EncryptionManager = None

try:
    from mautrix.crypto.attachments import encrypt_attachment
except ImportError as e:
    if __optional_imports__:
        raise
    media_encrypt_import_error = e
    encrypt_attachment = None

EVENT_TIME = Histogram(
    "bridge_matrix_event", "Time spent processing Matrix events", ["event_type"]
)


class UnencryptedMessageError(DecryptionError):
    def __init__(self) -> None:
        super().__init__("unencrypted message")

    @property
    def human_message(self) -> str:
        return "the message is not encrypted"


class EncryptionUnsupportedError(DecryptionError):
    def __init__(self) -> None:
        super().__init__("encryption is not supported")

    @property
    def human_message(self) -> str:
        return "the bridge is not configured to support encryption"


class DeviceUntrustedError(DecryptionError):
    def __init__(self, trust: TrustState) -> None:
        explanation = {
            TrustState.BLACKLISTED: "device is blacklisted",
            TrustState.UNVERIFIED: "unverified",
            TrustState.UNKNOWN_DEVICE: "device info not found",
            TrustState.FORWARDED: "keys were forwarded from an unknown device",
            TrustState.CROSS_SIGNED_UNTRUSTED: (
                "cross-signing keys changed after setting up the bridge"
            ),
        }.get(trust)
        base = "your device is not trusted"
        self.message = f"{base} ({explanation})" if explanation else base
        super().__init__(self.message)

    @property
    def human_message(self) -> str:
        return self.message


class BaseMatrixHandler:
    log: TraceLogger = logging.getLogger("mau.mx")
    az: AppService
    commands: cmd.CommandProcessor
    config: config.BaseBridgeConfig
    bridge: br.Bridge
    e2ee: EncryptionManager | None
    require_e2ee: bool
    media_config: MediaRepoConfig
    versions: VersionsResponse
    minimum_spec_version: Version = SpecVersions.V11
    room_locks: dict[str, asyncio.Lock]

    user_id_prefix: str
    user_id_suffix: str

    def __init__(
        self,
        command_processor: cmd.CommandProcessor | None = None,
        bridge: br.Bridge | None = None,
    ) -> None:
        self.az = bridge.az
        self.config = bridge.config
        self.bridge = bridge
        self.commands = command_processor or cmd.CommandProcessor(bridge=bridge)
        self.media_config = MediaRepoConfig(upload_size=50 * 1024 * 1024)
        self.versions = VersionsResponse.deserialize({"versions": ["v1.3"]})
        self.az.matrix_event_handler(self.int_handle_event)
        self.room_locks = defaultdict(asyncio.Lock)

        self.e2ee = None
        self.require_e2ee = False
        if self.config["bridge.encryption.allow"]:
            if not EncryptionManager:
                self.log.fatal(
                    "Encryption enabled in config, but dependencies not installed.",
                    exc_info=encryption_import_error,
                )
                sys.exit(31)
            if not encrypt_attachment:
                self.log.fatal(
                    "Encryption enabled in config, but media encryption dependencies "
                    "not installed.",
                    exc_info=media_encrypt_import_error,
                )
                sys.exit(31)
            self.e2ee = EncryptionManager(
                bridge=bridge,
                user_id_prefix=self.user_id_prefix,
                user_id_suffix=self.user_id_suffix,
                homeserver_address=self.config["homeserver.address"],
                db_url=self.config["appservice.database"],
            )
            self.require_e2ee = self.config["bridge.encryption.require"]

        self.management_room_text = self.config.get(
            "bridge.management_room_text",
            {
                "welcome": "Hello, I'm a bridge bot.",
                "welcome_connected": "Use `help` for help.",
                "welcome_unconnected": "Use `help` for help on how to log in.",
            },
        )
        self.management_room_multiple_messages = self.config.get(
            "bridge.management_room_multiple_messages",
            False,
        )

    async def check_versions(self) -> None:
        if not self.versions.supports_at_least(self.minimum_spec_version):
            self.log.fatal(
                "The homeserver is outdated "
                "(server supports Matrix %s, but the bridge requires at least %s)",
                self.versions.latest_version,
                self.minimum_spec_version,
            )
            sys.exit(18)
        if self.bridge.homeserver_software.is_hungry and not self.versions.supports(
            "com.beeper.hungry"
        ):
            self.log.fatal(
                "The config claims the homeserver is hungryserv, "
                "but the /versions response didn't confirm it"
            )
            sys.exit(18)

    async def wait_for_connection(self) -> None:
        self.log.info("Ensuring connectivity to homeserver")
        while True:
            try:
                self.versions = await self.az.intent.versions()
                break
            except Exception:
                self.log.exception("Connection to homeserver failed, retrying in 10 seconds")
                await asyncio.sleep(10)
        await self.check_versions()
        try:
            await self.az.intent.whoami()
        except MForbidden:
            self.log.debug(
                "Whoami endpoint returned M_FORBIDDEN, "
                "trying to register bridge bot before retrying..."
            )
            await self.az.intent.ensure_registered()
            await self.az.intent.whoami()
        if self.versions.supports("fi.mau.msc2659.stable") or self.versions.supports_at_least(
            SpecVersions.V17
        ):
            try:
                txn_id = self.az.intent.api.get_txn_id()
                duration = await self.az.ping_self(txn_id)
                self.log.debug(
                    "Homeserver->bridge connection works, "
                    f"roundtrip time is {duration} ms (txn ID: {txn_id})"
                )
            except Exception:
                self.log.exception("Error checking homeserver -> bridge connection")
                sys.exit(16)
        else:
            self.log.debug(
                "Homeserver does not support checking status of homeserver -> bridge connection"
            )
        try:
            self.media_config = await self.az.intent.get_media_repo_config()
        except Exception:
            self.log.warning("Failed to fetch media repo config", exc_info=True)

    async def init_as_bot(self) -> None:
        self.log.debug("Initializing appservice bot")
        displayname = self.config["appservice.bot_displayname"]
        if displayname:
            try:
                await self.az.intent.set_displayname(
                    displayname if displayname != "remove" else ""
                )
            except Exception:
                self.log.exception("Failed to set bot displayname")

        avatar = self.config["appservice.bot_avatar"]
        if avatar:
            try:
                await self.az.intent.set_avatar_url(avatar if avatar != "remove" else "")
            except Exception:
                self.log.exception("Failed to set bot avatar")

        if self.bridge.homeserver_software.is_hungry and self.bridge.beeper_network_name:
            self.log.debug("Setting contact info on the appservice bot")
            await self.az.intent.beeper_update_profile(
                {
                    "com.beeper.bridge.service": self.bridge.beeper_service_name,
                    "com.beeper.bridge.network": self.bridge.beeper_network_name,
                    "com.beeper.bridge.is_bridge_bot": True,
                }
            )

    async def init_encryption(self) -> None:
        if self.e2ee:
            await self.e2ee.start()

    async def allow_message(self, user: br.BaseUser) -> bool:
        return user.is_whitelisted or (
            self.config["bridge.relay.enabled"] and user.relay_whitelisted
        )

    @staticmethod
    async def allow_command(user: br.BaseUser) -> bool:
        return user.is_whitelisted

    @staticmethod
    async def allow_bridging_message(user: br.BaseUser, portal: br.BasePortal) -> bool:
        return await user.is_logged_in() or (user.relay_whitelisted and portal.has_relay)

    @staticmethod
    async def allow_puppet_invite(user: br.BaseUser, puppet: br.BasePuppet) -> bool:
        return await user.is_logged_in()

    async def handle_leave(self, room_id: RoomID, user_id: UserID, event_id: EventID) -> None:
        pass

    async def handle_kick(
        self, room_id: RoomID, user_id: UserID, kicked_by: UserID, reason: str, event_id: EventID
    ) -> None:
        pass

    async def handle_ban(
        self, room_id: RoomID, user_id: UserID, banned_by: UserID, reason: str, event_id: EventID
    ) -> None:
        pass

    async def handle_unban(
        self, room_id: RoomID, user_id: UserID, unbanned_by: UserID, reason: str, event_id: EventID
    ) -> None:
        pass

    async def handle_join(self, room_id: RoomID, user_id: UserID, event_id: EventID) -> None:
        pass

    async def handle_knock(
        self, room_id: RoomID, user_id: UserID, reason: str, event_id: EventID
    ) -> None:
        pass

    async def handle_retract_knock(
        self, room_id: RoomID, user_id: UserID, reason: str, event_id: EventID
    ) -> None:
        pass

    async def handle_reject_knock(
        self, room_id: RoomID, user_id: UserID, sender: UserID, reason: str, event_id: EventID
    ) -> None:
        pass

    async def handle_accept_knock(
        self, room_id: RoomID, user_id: UserID, sender: UserID, reason: str, event_id: EventID
    ) -> None:
        pass

    async def handle_member_info_change(
        self,
        room_id: RoomID,
        user_id: UserID,
        content: MemberStateEventContent,
        prev_content: MemberStateEventContent,
        event_id: EventID,
    ) -> None:
        pass

    async def handle_puppet_group_invite(
        self,
        room_id: RoomID,
        puppet: br.BasePuppet,
        invited_by: br.BaseUser,
        evt: StateEvent,
        members: list[UserID],
    ) -> None:
        if self.az.bot_mxid not in members:
            await puppet.default_mxid_intent.leave_room(
                room_id, reason="This ghost does not join multi-user rooms without the bridge bot."
            )

    async def handle_puppet_dm_invite(
        self, room_id: RoomID, puppet: br.BasePuppet, invited_by: br.BaseUser, evt: StateEvent
    ) -> None:
        portal = await invited_by.get_portal_with(puppet)
        if portal:
            await portal.accept_matrix_dm(room_id, invited_by, puppet)
        else:
            await puppet.default_mxid_intent.leave_room(
                room_id, reason="This bridge does not support creating DMs."
            )

    async def handle_puppet_space_invite(
        self, room_id: RoomID, puppet: br.BasePuppet, invited_by: br.BaseUser, evt: StateEvent
    ) -> None:
        await puppet.default_mxid_intent.leave_room(
            room_id, reason="This ghost does not join spaces."
        )

    async def handle_puppet_nonportal_invite(
        self, room_id: RoomID, puppet: br.BasePuppet, invited_by: br.BaseUser, evt: StateEvent
    ) -> None:
        intent = puppet.default_mxid_intent
        await intent.join_room(room_id)
        try:
            create_evt = await intent.get_state_event(room_id, EventType.ROOM_CREATE)
            members = await intent.get_room_members(room_id)
        except MatrixError:
            self.log.exception(f"Failed to get state after joining {room_id} as {intent.mxid}")
            background_task.create(intent.leave_room(room_id, reason="Internal error"))
            return
        if create_evt.type == RoomType.SPACE:
            await self.handle_puppet_space_invite(room_id, puppet, invited_by, evt)
        elif len(members) > 2 or not evt.content.is_direct:
            await self.handle_puppet_group_invite(room_id, puppet, invited_by, evt, members)
        else:
            await self.handle_puppet_dm_invite(room_id, puppet, invited_by, evt)

    async def handle_puppet_invite(
        self, room_id: RoomID, puppet: br.BasePuppet, invited_by: br.BaseUser, evt: StateEvent
    ) -> None:
        intent = puppet.default_mxid_intent
        if not await self.allow_puppet_invite(invited_by, puppet):
            self.log.debug(f"Rejecting invite for {intent.mxid} to {room_id}: user can't invite")
            await intent.leave_room(room_id, reason="You're not allowed to invite this ghost.")
            return

        async with self.room_locks[room_id]:
            portal = await self.bridge.get_portal(room_id)
            if portal:
                try:
                    await portal.handle_matrix_invite(invited_by, puppet)
                except br.RejectMatrixInvite as e:
                    await intent.leave_room(room_id, reason=e.message)
                except br.IgnoreMatrixInvite:
                    pass
                else:
                    await intent.join_room(room_id)
                return
            else:
                await self.handle_puppet_nonportal_invite(room_id, puppet, invited_by, evt)

    async def handle_invite(
        self, room_id: RoomID, user_id: UserID, invited_by: br.BaseUser, evt: StateEvent
    ) -> None:
        pass

    async def handle_reject(
        self, room_id: RoomID, user_id: UserID, reason: str, event_id: EventID
    ) -> None:
        pass

    async def handle_disinvite(
        self,
        room_id: RoomID,
        user_id: UserID,
        disinvited_by: UserID,
        reason: str,
        event_id: EventID,
    ) -> None:
        pass

    async def handle_event(self, evt: Event) -> None:
        """
        Called by :meth:`int_handle_event` for message events other than m.room.message.

        **N.B.** You may need to add the event class to :attr:`allowed_event_classes`
        or override :meth:`allow_matrix_event` for it to reach here.
        """

    async def handle_state_event(self, evt: StateEvent) -> None:
        """
        Called by :meth:`int_handle_event` for state events other than m.room.membership.

        **N.B.** You may need to add the event class to :attr:`allowed_event_classes`
        or override :meth:`allow_matrix_event` for it to reach here.
        """

    async def handle_ephemeral_event(
        self, evt: ReceiptEvent | PresenceEvent | TypingEvent
    ) -> None:
        if evt.type == EventType.RECEIPT:
            await self.handle_receipt(evt)

    async def send_permission_error(self, room_id: RoomID) -> None:
        await self.az.intent.send_notice(
            room_id,
            text=(
                "You are not whitelisted to use this bridge.\n\n"
                "If you are the owner of this bridge, see the bridge.permissions "
                "section in your config file."
            ),
            html=(
                "<p>You are not whitelisted to use this bridge.</p>"
                "<p>If you are the owner of this bridge, see the "
                "<code>bridge.permissions</code> section in your config file.</p>"
            ),
        )

    async def accept_bot_invite(self, room_id: RoomID, inviter: br.BaseUser) -> None:
        try:
            await self.az.intent.join_room(room_id)
        except Exception:
            self.log.exception(f"Failed to join room {room_id} as bridge bot")
            return

        if not await self.allow_command(inviter):
            await self.send_permission_error(room_id)
            await self.az.intent.leave_room(room_id)
            return

        await self.send_welcome_message(room_id, inviter)

    async def send_welcome_message(self, room_id: RoomID, inviter: br.BaseUser) -> None:
        has_two_members, bridge_bot_in_room = await self._is_direct_chat(room_id)
        is_management = has_two_members and bridge_bot_in_room

        welcome_messages = [self.management_room_text.get("welcome")]

        if is_management:
            if await inviter.is_logged_in():
                welcome_messages.append(self.management_room_text.get("welcome_connected"))
            else:
                welcome_messages.append(self.management_room_text.get("welcome_unconnected"))

            additional_help = self.management_room_text.get("additional_help")
            if additional_help:
                welcome_messages.append(additional_help)
        else:
            cmd_prefix = self.commands.command_prefix
            welcome_messages.append(f"Use `{cmd_prefix} help` for help.")

        if self.management_room_multiple_messages:
            for m in welcome_messages:
                await self.az.intent.send_notice(room_id, text=m, html=markdown.render(m))
        else:
            combined = "\n".join(welcome_messages)
            combined_html = "".join(map(markdown.render, welcome_messages))
            await self.az.intent.send_notice(room_id, text=combined, html=combined_html)

    async def int_handle_invite(self, evt: StateEvent) -> None:
        self.log.debug(f"{evt.sender} invited {evt.state_key} to {evt.room_id}")
        inviter = await self.bridge.get_user(evt.sender)
        if inviter is None:
            self.log.exception(f"Failed to find user with Matrix ID {evt.sender}")
            return
        elif evt.state_key == self.az.bot_mxid:
            await self.accept_bot_invite(evt.room_id, inviter)
            return

        puppet = await self.bridge.get_puppet(UserID(evt.state_key))
        if puppet:
            await self.handle_puppet_invite(evt.room_id, puppet, inviter, evt)
            return

        await self.handle_invite(evt.room_id, UserID(evt.state_key), inviter, evt)

    def is_command(self, message: MessageEventContent) -> tuple[bool, str]:
        text = message.body
        prefix = self.config["bridge.command_prefix"]
        is_command = text.startswith(prefix)
        if is_command:
            text = text[len(prefix) + 1 :].lstrip()
        return is_command, text

    async def _send_mss(
        self,
        evt: Event,
        status: MessageStatus,
        reason: MessageStatusReason | None = None,
        error: str | None = None,
        message: str | None = None,
    ) -> None:
        if not self.config.get("bridge.message_status_events", False):
            return
        status_content = BeeperMessageStatusEventContent(
            network="",  # TODO set network properly
            relates_to=RelatesTo(rel_type=RelationType.REFERENCE, event_id=evt.event_id),
            status=status,
            reason=reason,
            error=error,
            message=message,
        )
        await self.az.intent.send_message_event(
            evt.room_id, EventType.BEEPER_MESSAGE_STATUS, status_content
        )

    async def _send_crypto_status_error(
        self,
        evt: Event,
        err: DecryptionError | None = None,
        retry_num: int = 0,
        is_final: bool = True,
        edit: EventID | None = None,
        wait_for: int | None = None,
    ) -> EventID | None:
        msg = str(err)
        if isinstance(err, (SessionNotFound, UnencryptedMessageError)):
            msg = err.human_message
        self._send_message_checkpoint(
            evt, MessageSendCheckpointStep.DECRYPTED, msg, permanent=is_final, retry_num=retry_num
        )

        if wait_for:
            msg += f". The bridge will retry for {wait_for} seconds"
        full_msg = f"\u26a0 Your message was not bridged: {msg}."
        if isinstance(err, EncryptionUnsupportedError):
            full_msg = "🔒️ This bridge has not been configured to support encryption"
        event_id = None
        if self.config.get("bridge.delivery_error_reports", True):
            try:
                content = TextMessageEventContent(msgtype=MessageType.NOTICE, body=full_msg)
                if edit:
                    content.set_edit(edit)
                event_id = await self.az.intent.send_message(evt.room_id, content)
            except IntentError:
                self.log.debug("IntentError while sending encryption error", exc_info=True)
                self.log.error(
                    "Got IntentError while trying to send encryption error message. "
                    "This likely means the bridge bot is not in the room, which can "
                    "happen if you force-enable e2ee on the homeserver without enabling "
                    "it by default on the bridge (bridge -> encryption -> default)."
                )

        await self._send_mss(
            evt,
            status=MessageStatus.RETRIABLE if is_final else MessageStatus.PENDING,
            reason=MessageStatusReason.UNDECRYPTABLE,
            error=msg,
            message=err.human_message if err else None,
        )

        return event_id

    async def handle_message(self, evt: MessageEvent, was_encrypted: bool = False) -> None:
        room_id = evt.room_id
        user_id = evt.sender
        event_id = evt.event_id
        message = evt.content

        if not was_encrypted and self.require_e2ee:
            self.log.warning(f"Dropping {event_id} from {user_id} as it's not encrypted!")
            await self._send_crypto_status_error(evt, UnencryptedMessageError(), 0)
            return

        sender = await self.bridge.get_user(user_id)
        if not sender or not await self.allow_message(sender):
            self.log.debug(
                f"Ignoring message {event_id} from {user_id} to {room_id}:"
                " user is not whitelisted."
            )
            self._send_message_checkpoint(
                evt, MessageSendCheckpointStep.BRIDGE, "user is not whitelisted"
            )
            return
        self.log.debug(f"Received Matrix event {event_id} from {sender.mxid} in {room_id}")
        self.log.trace("Event %s content: %s", event_id, message)

        if isinstance(message, TextMessageEventContent):
            message.trim_reply_fallback()

        is_command, text = self.is_command(message)
        portal = await self.bridge.get_portal(room_id)
        if not is_command and portal:
            if await self.allow_bridging_message(sender, portal):
                await portal.handle_matrix_message(sender, message, event_id)
            else:
                self.log.debug(
                    f"Ignoring event {event_id} from {sender.mxid}:"
                    " not allowed to send to portal"
                )
                self._send_message_checkpoint(
                    evt,
                    MessageSendCheckpointStep.BRIDGE,
                    "user is not allowed to send to the portal",
                )
            return

        if message.msgtype != MessageType.TEXT:
            self.log.debug(
                f"Ignoring event {event_id}: not a portal room and not a m.text message"
            )
            self._send_message_checkpoint(
                evt, MessageSendCheckpointStep.BRIDGE, "not a portal room and not a m.text message"
            )
            return
        elif not await self.allow_command(sender):
            self.log.debug(
                f"Ignoring command {event_id} from {sender.mxid}: not allowed to run commands"
            )
            self._send_message_checkpoint(
                evt, MessageSendCheckpointStep.COMMAND, "not allowed to run commands"
            )
            return

        has_two_members, bridge_bot_in_room = await self._is_direct_chat(room_id)
        is_management = has_two_members and bridge_bot_in_room

        if is_command or is_management:
            try:
                command, arguments = text.split(" ", 1)
                args = arguments.split(" ")
            except ValueError:
                # Not enough values to unpack, i.e. no arguments
                command = text
                args = []

            try:
                await self.commands.handle(
                    room_id,
                    event_id,
                    sender,
                    command,
                    args,
                    message,
                    portal,
                    is_management,
                    bridge_bot_in_room,
                )
            except Exception as e:
                self.log.debug(f"Error handling command {command} from {sender}: {e}")
                self._send_message_checkpoint(evt, MessageSendCheckpointStep.COMMAND, e)
                await self._send_mss(
                    evt,
                    status=MessageStatus.FAIL,
                    reason=MessageStatusReason.GENERIC_ERROR,
                    error="",
                    message="Command execution failed",
                )
            else:
                await MessageSendCheckpoint(
                    event_id=event_id,
                    room_id=room_id,
                    step=MessageSendCheckpointStep.COMMAND,
                    timestamp=int(time.time() * 1000),
                    status=MessageSendCheckpointStatus.SUCCESS,
                    reported_by=MessageSendCheckpointReportedBy.BRIDGE,
                    event_type=EventType.ROOM_MESSAGE,
                    message_type=message.msgtype,
                ).send(
                    self.bridge.config["homeserver.message_send_checkpoint_endpoint"],
                    self.az.as_token,
                    self.log,
                )
                await self._send_mss(evt, status=MessageStatus.SUCCESS)
        else:
            self.log.debug(
                f"Ignoring event {event_id} from {sender.mxid}:"
                " not a command and not a portal room"
            )
            self._send_message_checkpoint(
                evt, MessageSendCheckpointStep.COMMAND, "not a command and not a portal room"
            )
            await self._send_mss(
                evt,
                status=MessageStatus.FAIL,
                reason=MessageStatusReason.UNSUPPORTED,
                error="Unknown room",
                message="Unknown room",
            )

    async def _is_direct_chat(self, room_id: RoomID) -> tuple[bool, bool]:
        try:
            members = await self.az.intent.get_room_members(room_id)
            return len(members) == 2, self.az.bot_mxid in members
        except MatrixError:
            return False, False

    async def handle_receipt(self, evt: ReceiptEvent) -> None:
        for event_id, receipts in evt.content.items():
            for user_id, data in receipts.get(ReceiptType.READ, {}).items():
                user = await self.bridge.get_user(user_id, create=False)
                if not user or not await user.is_logged_in():
                    continue

                portal = await self.bridge.get_portal(evt.room_id)
                if not portal:
                    continue

                await portal.schedule_disappearing()

                if (
                    data.get(DOUBLE_PUPPET_SOURCE_KEY) == self.az.bridge_name
                    and await self.bridge.get_double_puppet(user_id) is not None
                ):
                    continue

                await self.handle_read_receipt(user, portal, event_id, data)

    async def handle_read_receipt(
        self,
        user: br.BaseUser,
        portal: br.BasePortal,
        event_id: EventID,
        data: SingleReceiptEventContent,
    ) -> None:
        pass

    async def try_handle_sync_event(self, evt: Event) -> None:
        try:
            if isinstance(evt, (ReceiptEvent, PresenceEvent, TypingEvent)):
                await self.handle_ephemeral_event(evt)
            else:
                self.log.trace("Unknown event type received from sync: %s", evt)
        except Exception:
            self.log.exception("Error handling manually received Matrix event")

    async def _post_decrypt(
        self, evt: Event, retry_num: int = 0, error_event_id: EventID | None = None
    ) -> None:
        trust_state = evt["mautrix"]["trust_state"]
        if trust_state < self.e2ee.min_send_trust:
            self.log.warning(
                f"Dropping {evt.event_id} from {evt.sender} due to insufficient verification level"
                f" (event: {trust_state}, required: {self.e2ee.min_send_trust})"
            )
            await self._send_crypto_status_error(
                evt,
                retry_num=retry_num,
                err=DeviceUntrustedError(trust_state),
                edit=error_event_id,
            )
            return

        self._send_message_checkpoint(
            evt, MessageSendCheckpointStep.DECRYPTED, retry_num=retry_num
        )
        if error_event_id:
            await self.az.intent.redact(evt.room_id, error_event_id)
        await self.int_handle_event(evt, was_encrypted=True)

    async def handle_encrypted(self, evt: EncryptedEvent) -> None:
        if not self.e2ee:
            self.log.debug(
                "Got encrypted message %s from %s, but encryption is not enabled",
                evt.event_id,
                evt.sender,
            )
            await self._send_crypto_status_error(evt, EncryptionUnsupportedError())
            return
        try:
            decrypted = await self.e2ee.decrypt(evt, wait_session_timeout=3)
        except SessionNotFound as e:
            await self._handle_encrypted_wait(evt, e, wait=22)
        except DecryptionError as e:
            self.log.warning(f"Failed to decrypt {evt.event_id}: {e}")
            self.log.trace("%s decryption traceback:", evt.event_id, exc_info=True)
            await self._send_crypto_status_error(evt, e)
        else:
            await self._post_decrypt(decrypted)

    async def _handle_encrypted_wait(
        self, evt: EncryptedEvent, err: SessionNotFound, wait: int
    ) -> None:
        self.log.debug(
            f"Couldn't find session {err.session_id} trying to decrypt {evt.event_id},"
            " waiting even longer"
        )
        background_task.create(
            self.e2ee.crypto.request_room_key(
                evt.room_id,
                evt.content.sender_key,
                evt.content.session_id,
                from_devices={evt.sender: [evt.content.device_id]},
            )
        )
        event_id = await self._send_crypto_status_error(evt, err, is_final=False, wait_for=wait)
        got_keys = await self.e2ee.crypto.wait_for_session(
            evt.room_id, err.session_id, timeout=wait
        )
        if got_keys:
            self.log.debug(
                f"Got session {err.session_id} after waiting more, "
                f"trying to decrypt {evt.event_id} again"
            )
            try:
                decrypted = await self.e2ee.decrypt(evt, wait_session_timeout=0)
            except DecryptionError as e:
                await self._send_crypto_status_error(evt, e, retry_num=1, edit=event_id)
                self.log.warning(f"Failed to decrypt {evt.event_id}: {e}")
                self.log.trace("%s decryption traceback:", evt.event_id, exc_info=True)
            else:
                await self._post_decrypt(decrypted, retry_num=1, error_event_id=event_id)
                return
        else:
            self.log.warning(f"Didn't get {err.session_id}, giving up on {evt.event_id}")
            await self._send_crypto_status_error(
                evt, SessionNotFound(err.session_id), retry_num=1, edit=event_id
            )

    async def handle_encryption(self, evt: StateEvent) -> None:
        await self.az.state_store.set_encryption_info(evt.room_id, evt.content)
        portal = await self.bridge.get_portal(evt.room_id)
        if portal:
            portal.encrypted = True
            await portal.save()
            if portal.is_direct:
                portal.log.debug("Received encryption event in direct portal: %s", evt.content)
                await portal.enable_dm_encryption()

    def _send_message_checkpoint(
        self,
        evt: Event,
        step: MessageSendCheckpointStep,
        err: Exception | str | None = None,
        permanent: bool = True,
        retry_num: int = 0,
    ) -> None:
        endpoint = self.bridge.config["homeserver.message_send_checkpoint_endpoint"]
        if not endpoint:
            return
        if evt.type not in CHECKPOINT_TYPES:
            return

        self.log.debug(f"Sending message send checkpoint for {evt.event_id} (step: {step})")
        status = MessageSendCheckpointStatus.SUCCESS
        if err:
            status = (
                MessageSendCheckpointStatus.PERM_FAILURE
                if permanent
                else MessageSendCheckpointStatus.WILL_RETRY
            )

        checkpoint = MessageSendCheckpoint(
            event_id=evt.event_id,
            room_id=evt.room_id,
            step=step,
            timestamp=int(time.time() * 1000),
            status=status,
            reported_by=MessageSendCheckpointReportedBy.BRIDGE,
            event_type=evt.type,
            message_type=evt.content.msgtype if evt.type == EventType.ROOM_MESSAGE else None,
            info=str(err) if err else None,
            retry_num=retry_num,
        )
        background_task.create(checkpoint.send(endpoint, self.az.as_token, self.log))

    allowed_event_classes: tuple[type, ...] = (
        MessageEvent,
        StateEvent,
        ReactionEvent,
        EncryptedEvent,
        RedactionEvent,
        ReceiptEvent,
        TypingEvent,
        PresenceEvent,
    )

    async def allow_matrix_event(self, evt: Event) -> bool:
        # If the event is not one of the allowed classes, ignore it.
        if not isinstance(evt, self.allowed_event_classes):
            return False
        # For room events, make sure the message didn't originate from the bridge.
        if isinstance(evt, BaseRoomEvent):
            # If the event is from a bridge ghost, ignore it.
            if evt.sender == self.az.bot_mxid or self.bridge.is_bridge_ghost(evt.sender):
                return False
            # If the event is marked as double puppeted and we can confirm that we are in fact
            # double puppeting that user ID, ignore it.
            if (
                evt.content.get(DOUBLE_PUPPET_SOURCE_KEY) == self.az.bridge_name
                and await self.bridge.get_double_puppet(evt.sender) is not None
            ):
                return False
        # For non-room events and non-bridge-originated room events, allow.
        return True

    async def int_handle_event(self, evt: Event, was_encrypted: bool = False) -> None:
        if isinstance(evt, StateEvent) and evt.type == EventType.ROOM_MEMBER and self.e2ee:
            await self.e2ee.handle_member_event(evt)
        if not await self.allow_matrix_event(evt):
            return
        self.log.trace("Received event: %s", evt)

        if not was_encrypted:
            self._send_message_checkpoint(evt, MessageSendCheckpointStep.BRIDGE)
        start_time = time.time()

        if evt.type == EventType.ROOM_MEMBER:
            evt: StateEvent
            unsigned = evt.unsigned or StateUnsigned()
            prev_content = unsigned.prev_content or MemberStateEventContent()
            prev_membership = prev_content.membership if prev_content else Membership.JOIN
            if evt.content.membership == Membership.INVITE:
                if prev_membership == Membership.KNOCK:
                    await self.handle_accept_knock(
                        evt.room_id,
                        UserID(evt.state_key),
                        evt.sender,
                        evt.content.reason,
                        evt.event_id,
                    )
                else:
                    await self.int_handle_invite(evt)
            elif evt.content.membership == Membership.LEAVE:
                if prev_membership == Membership.BAN:
                    await self.handle_unban(
                        evt.room_id,
                        UserID(evt.state_key),
                        evt.sender,
                        evt.content.reason,
                        evt.event_id,
                    )
                elif prev_membership == Membership.INVITE:
                    if evt.sender == evt.state_key:
                        await self.handle_reject(
                            evt.room_id, UserID(evt.state_key), evt.content.reason, evt.event_id
                        )
                    else:
                        await self.handle_disinvite(
                            evt.room_id,
                            UserID(evt.state_key),
                            evt.sender,
                            evt.content.reason,
                            evt.event_id,
                        )
                elif prev_membership == Membership.KNOCK:
                    if evt.sender == evt.state_key:
                        await self.handle_retract_knock(
                            evt.room_id, UserID(evt.state_key), evt.content.reason, evt.event_id
                        )
                    else:
                        await self.handle_reject_knock(
                            evt.room_id,
                            UserID(evt.state_key),
                            evt.sender,
                            evt.content.reason,
                            evt.event_id,
                        )

                elif evt.sender == evt.state_key:
                    await self.handle_leave(evt.room_id, UserID(evt.state_key), evt.event_id)
                else:
                    await self.handle_kick(
                        evt.room_id,
                        UserID(evt.state_key),
                        evt.sender,
                        evt.content.reason,
                        evt.event_id,
                    )
            elif evt.content.membership == Membership.BAN:
                await self.handle_ban(
                    evt.room_id,
                    UserID(evt.state_key),
                    evt.sender,
                    evt.content.reason,
                    evt.event_id,
                )
            elif evt.content.membership == Membership.JOIN:
                if prev_membership != Membership.JOIN:
                    await self.handle_join(evt.room_id, UserID(evt.state_key), evt.event_id)
                else:
                    await self.handle_member_info_change(
                        evt.room_id, UserID(evt.state_key), evt.content, prev_content, evt.event_id
                    )
            elif evt.content.membership == Membership.KNOCK:
                await self.handle_knock(
                    evt.room_id,
                    UserID(evt.state_key),
                    evt.content.reason,
                    evt.event_id,
                )
        elif evt.type in (EventType.ROOM_MESSAGE, EventType.STICKER):
            evt: MessageEvent
            if evt.type != EventType.ROOM_MESSAGE:
                evt.content.msgtype = MessageType(str(evt.type))
            await self.handle_message(evt, was_encrypted=was_encrypted)
        elif evt.type == EventType.ROOM_ENCRYPTED:
            await self.handle_encrypted(evt)
        elif evt.type == EventType.ROOM_ENCRYPTION:
            await self.handle_encryption(evt)
        else:
            if evt.type.is_state and isinstance(evt, StateEvent):
                await self.handle_state_event(evt)
            elif evt.type.is_ephemeral and isinstance(
                evt, (PresenceEvent, TypingEvent, ReceiptEvent)
            ):
                await self.handle_ephemeral_event(evt)
            else:
                await self.handle_event(evt)

        await self.log_event_handle_duration(evt, time.time() - start_time)

    async def log_event_handle_duration(self, evt: Event, duration: float) -> None:
        EVENT_TIME.labels(event_type=str(evt.type)).observe(duration)



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import NamedTuple

import pytest

from .matrixuri import IdentifierType, MatrixURI, MatrixURIError, URIAction, _PathPart
from .primitive import EventID, RoomAlias, RoomID, UserID


def test_basic_parse_uri() -> None:
    for test in basic_tests:
        assert MatrixURI.parse(test.uri) == test.parsed


def test_basic_stringify_uri() -> None:
    for test in basic_tests:
        assert test.uri == test.parsed.matrix_uri


def test_basic_parse_url() -> None:
    for test in basic_tests:
        assert MatrixURI.parse(test.url) == test.parsed


def test_basic_stringify_url() -> None:
    for test in basic_tests:
        assert test.url == test.parsed.matrix_to_url


def test_basic_build() -> None:
    for test in basic_tests:
        assert MatrixURI.build(*test.params) == test.parsed


def test_parse_unescaped() -> None:
    assert MatrixURI.parse("https://matrix.to/#/#hello:world").room_alias == "#hello:world"


def test_parse_trailing_slash() -> None:
    assert MatrixURI.parse("https://matrix.to/#/#hello:world/").room_alias == "#hello:world"
    assert MatrixURI.parse("matrix:r/hello:world/").room_alias == "#hello:world"


def test_parse_errors() -> None:
    tests = [
        "https://example.com",
        "matrix:invalid/foo",
        "matrix:hello world",
        "matrix:/roomid",
        "matrix:roomid/",
        "matrix:roomid/foo/e/",
        "matrix:roomid/foo/e",
        "https://matrix.to",
        "https://matrix.to/#/",
        "https://matrix.to/#foo/#hello:world",
        "https://matrix.to/#/#hello:world/hmm",
    ]
    for test in tests:
        with pytest.raises(MatrixURIError):
            print(MatrixURI.parse(test))


def test_build_errors() -> None:
    with pytest.raises(ValueError):
        MatrixURI.build("hello world")
    with pytest.raises(ValueError):
        MatrixURI.build(EventID("$uOH4C9cK4HhMeFWkUXMbdF_dtndJ0j9je-kIK3XpV1s"))
    with pytest.raises(ValueError):
        MatrixURI.build(
            UserID("@user:example.org"),
            EventID("$uOH4C9cK4HhMeFWkUXMbdF_dtndJ0j9je-kIK3XpV1s"),
        )
    with pytest.raises(ValueError):
        MatrixURI.build(
            RoomID("!room:example.org"),
            RoomID("!anotherroom:example.com"),
        )
    with pytest.raises(ValueError):
        MatrixURI.build(
            RoomID("!room:example.org"),
            "hmm",
        )


def _make_parsed(
    part1: _PathPart,
    part2: _PathPart | None = None,
    via: list[str] | None = None,
    action: URIAction | None = None,
) -> MatrixURI:
    uri = MatrixURI()
    uri._part1 = part1
    uri._part2 = part2
    uri.via = via
    uri.action = action
    return uri


class BasicTestItems(NamedTuple):
    url: str
    uri: str
    parsed: MatrixURI
    params: tuple[RoomID | UserID | RoomAlias, EventID | None, list[str] | None, URIAction | None]


basic_tests = [
    BasicTestItems(
        "https://matrix.to/#/%217NdBVvkd4aLSbgKt9RXl%3Aexample.org",
        "matrix:roomid/7NdBVvkd4aLSbgKt9RXl:example.org",
        _make_parsed(_PathPart(IdentifierType.ROOM_ID, "7NdBVvkd4aLSbgKt9RXl:example.org")),
        (RoomID("!7NdBVvkd4aLSbgKt9RXl:example.org"), None, None, None),
    ),
    BasicTestItems(
        "https://matrix.to/#/%217NdBVvkd4aLSbgKt9RXl%3Aexample.org?via=maunium.net&via=matrix.org",
        "matrix:roomid/7NdBVvkd4aLSbgKt9RXl:example.org?via=maunium.net&via=matrix.org",
        _make_parsed(
            _PathPart(IdentifierType.ROOM_ID, "7NdBVvkd4aLSbgKt9RXl:example.org"),
            via=["maunium.net", "matrix.org"],
        ),
        (RoomID("!7NdBVvkd4aLSbgKt9RXl:example.org"), None, ["maunium.net", "matrix.org"], None),
    ),
    BasicTestItems(
        "https://matrix.to/#/%23someroom%3Aexample.org",
        "matrix:r/someroom:example.org",
        _make_parsed(_PathPart(IdentifierType.ROOM_ALIAS, "someroom:example.org")),
        (RoomAlias("#someroom:example.org"), None, None, None),
    ),
    BasicTestItems(
        "https://matrix.to/#/%217NdBVvkd4aLSbgKt9RXl%3Aexample.org/%24uOH4C9cK4HhMeFWkUXMbdF_dtndJ0j9je-kIK3XpV1s",
        "matrix:roomid/7NdBVvkd4aLSbgKt9RXl:example.org/e/uOH4C9cK4HhMeFWkUXMbdF_dtndJ0j9je-kIK3XpV1s",
        _make_parsed(
            _PathPart(IdentifierType.ROOM_ID, "7NdBVvkd4aLSbgKt9RXl:example.org"),
            _PathPart(IdentifierType.EVENT, "uOH4C9cK4HhMeFWkUXMbdF_dtndJ0j9je-kIK3XpV1s"),
        ),
        (
            RoomID("!7NdBVvkd4aLSbgKt9RXl:example.org"),
            EventID("$uOH4C9cK4HhMeFWkUXMbdF_dtndJ0j9je-kIK3XpV1s"),
            None,
            None,
        ),
    ),
    BasicTestItems(
        "https://matrix.to/#/%23someroom%3Aexample.org/%24uOH4C9cK4HhMeFWkUXMbdF_dtndJ0j9je-kIK3XpV1s",
        "matrix:r/someroom:example.org/e/uOH4C9cK4HhMeFWkUXMbdF_dtndJ0j9je-kIK3XpV1s",
        _make_parsed(
            _PathPart(IdentifierType.ROOM_ALIAS, "someroom:example.org"),
            _PathPart(IdentifierType.EVENT, "uOH4C9cK4HhMeFWkUXMbdF_dtndJ0j9je-kIK3XpV1s"),
        ),
        (
            RoomAlias("#someroom:example.org"),
            EventID("$uOH4C9cK4HhMeFWkUXMbdF_dtndJ0j9je-kIK3XpV1s"),
            None,
            None,
        ),
    ),
    BasicTestItems(
        "https://matrix.to/#/%40user%3Aexample.org",
        "matrix:u/user:example.org",
        _make_parsed(_PathPart(IdentifierType.USER, "user:example.org")),
        (UserID("@user:example.org"), None, None, None),
    ),
]



from .async_attachments import (
    async_encrypt_attachment,
    async_generator_from_data,
    async_inplace_encrypt_attachment,
)
from .attachments import (
    decrypt_attachment,
    encrypt_attachment,
    encrypted_attachment_generator,
    inplace_encrypt_attachment,
)

__all__ = [
    "async_encrypt_attachment",
    "async_generator_from_data",
    "async_inplace_encrypt_attachment",
    "decrypt_attachment",
    "encrypt_attachment",
    "encrypted_attachment_generator",
    "inplace_encrypt_attachment",
]



# Copyright (c) 2023 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import Coroutine
import asyncio
import logging

_tasks = set()
log = logging.getLogger("mau.background_task")


async def catch(coro: Coroutine, caller: str) -> None:
    try:
        await coro
    except Exception:
        log.exception(f"Uncaught error in background task (created in {caller})")


# Logger.findCaller finds the 3rd stack frame, so add an intermediate function
# to get the caller of create().
def _find_caller() -> tuple[str, int, str, None]:
    return log.findCaller()


def create(coro: Coroutine, *, name: str | None = None, catch_errors: bool = True) -> asyncio.Task:
    """
    Create a background asyncio task safely, ensuring a reference is kept until the task completes.
    It also catches and logs uncaught errors (unless disabled via the parameter).

    Args:
        coro: The coroutine to wrap in a task and execute.
        name: An optional name for the created task.
        catch_errors: Should the task be wrapped in a try-except block to log any uncaught errors?

    Returns:
        An asyncio Task object wrapping the given coroutine.
    """
    if catch_errors:
        try:
            file_name, line_number, function_name, _ = _find_caller()
            caller = f"{function_name} at {file_name}:{line_number}"
        except ValueError:
            caller = "unknown function"
        task = asyncio.create_task(catch(coro, caller), name=name)
    else:
        task = asyncio.create_task(coro, name=name)
    _tasks.add(task)
    task.add_done_callback(_tasks.discard)
    return task



from __future__ import annotations

from typing import Awaitable, Callable, TypeVar
import asyncio
import json
import logging
import time
import urllib.request

from aiohttp import ClientConnectionError
from yarl import URL

from mautrix.util.logging import TraceLogger

try:
    from aiohttp_socks import ProxyConnectionError, ProxyError, ProxyTimeoutError
except ImportError:

    class ProxyError(Exception):
        pass

    ProxyConnectionError = ProxyTimeoutError = ProxyError

RETRYABLE_PROXY_EXCEPTIONS = (
    ProxyError,
    ProxyTimeoutError,
    ProxyConnectionError,
    ClientConnectionError,
    ConnectionError,
    asyncio.TimeoutError,
)


class ProxyHandler:
    current_proxy_url: str | None = None
    log = logging.getLogger("mau.proxy")

    def __init__(self, api_url: str | None) -> None:
        self.api_url = api_url

    def get_proxy_url_from_api(self, reason: str | None = None) -> str | None:
        assert self.api_url is not None

        api_url = str(URL(self.api_url).update_query({"reason": reason} if reason else {}))

        # NOTE: using urllib.request to intentionally block the whole bridge until the proxy change applied
        request = urllib.request.Request(api_url, method="GET")
        self.log.debug("Requesting proxy from: %s", api_url)

        try:
            with urllib.request.urlopen(request) as f:
                response = json.loads(f.read().decode())
        except Exception:
            self.log.exception("Failed to retrieve proxy from API")
            return self.current_proxy_url
        else:
            return response["proxy_url"]

    def update_proxy_url(self, reason: str | None = None) -> bool:
        old_proxy = self.current_proxy_url
        new_proxy = None

        if self.api_url is not None:
            new_proxy = self.get_proxy_url_from_api(reason)
        else:
            new_proxy = urllib.request.getproxies().get("http")

        if old_proxy != new_proxy:
            self.log.debug("Set new proxy URL: %s", new_proxy)
            self.current_proxy_url = new_proxy
            return True

        self.log.debug("Got same proxy URL: %s", new_proxy)
        return False

    def get_proxy_url(self) -> str | None:
        if not self.current_proxy_url:
            self.update_proxy_url()

        return self.current_proxy_url


T = TypeVar("T")


async def proxy_with_retry(
    name: str,
    func: Callable[[], Awaitable[T]],
    logger: TraceLogger,
    proxy_handler: ProxyHandler,
    on_proxy_change: Callable[[], Awaitable[None]],
    max_retries: int = 10,
    min_wait_seconds: int = 0,
    max_wait_seconds: int = 60,
    multiply_wait_seconds: int = 10,
    retryable_exceptions: tuple[Exception] = RETRYABLE_PROXY_EXCEPTIONS,
    reset_after_seconds: int | None = None,
) -> T:
    errors = 0
    last_error = 0

    while True:
        try:
            return await func()
        except retryable_exceptions as e:
            errors += 1
            if errors > max_retries:
                raise
            wait = errors * multiply_wait_seconds
            wait = max(wait, min_wait_seconds)
            wait = min(wait, max_wait_seconds)
            logger.warning(
                "%s while trying to %s, retrying in %d seconds",
                e.__class__.__name__,
                name,
                wait,
            )
            if errors > 1 and proxy_handler.update_proxy_url(
                f"{e.__class__.__name__} while trying to {name}"
            ):
                await on_proxy_change()

            # If sufficient time has passed since the previous error, reset the
            # error count. Useful for long running tasks with rare failures.
            if reset_after_seconds is not None:
                now = time.time()
                if last_error and now - last_error > reset_after_seconds:
                    errors = 0
                last_error = now



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from abc import ABC
import logging
import os
import pkgutil
import tempfile

from ruamel.yaml import YAML
from ruamel.yaml.comments import CommentedMap
from yarl import URL

from .base import BaseConfig
from .recursive_dict import RecursiveDict

yaml = YAML()
yaml.indent(4)
yaml.width = 200

log: logging.Logger = logging.getLogger("mau.util.config")


class BaseFileConfig(BaseConfig, ABC):
    def __init__(self, path: str, base_path: str) -> None:
        super().__init__()
        self._data = CommentedMap()
        self.path: str = path
        self.base_path: str = base_path

    def load(self) -> None:
        with open(self.path, "r") as stream:
            self._data = yaml.load(stream)

    def load_base(self) -> RecursiveDict[CommentedMap] | None:
        if self.base_path.startswith("pkg://"):
            url = URL(self.base_path)
            return RecursiveDict(yaml.load(pkgutil.get_data(url.host, url.path)), CommentedMap)
        try:
            with open(self.base_path, "r") as stream:
                return RecursiveDict(yaml.load(stream), CommentedMap)
        except OSError:
            pass
        return None

    def save(self) -> None:
        try:
            tf = tempfile.NamedTemporaryFile(
                mode="w", delete=False, suffix=".yaml", dir=os.path.dirname(self.path)
            )
        except OSError as e:
            log.warning(f"Failed to create tempfile to write updated config to disk: {e}")
            return
        try:
            yaml.dump(self._data, tf)
        except OSError as e:
            log.warning(f"Failed to write updated config to tempfile: {e}")
            tf.file.close()
            os.remove(tf.name)
            return
        tf.file.close()
        try:
            os.rename(tf.name, self.path)
        except OSError as e:
            log.warning(f"Failed to rename tempfile with updated config to {self.path}: {e}")
            try:
                os.remove(tf.name)
            except FileNotFoundError:
                pass



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from .authentication import ClientAuthenticationMethods
from .events import EventMethods
from .filtering import FilteringMethods
from .modules import ModuleMethods
from .rooms import RoomMethods
from .user_data import UserDataMethods


class ClientAPI(
    ClientAuthenticationMethods,
    FilteringMethods,
    RoomMethods,
    EventMethods,
    UserDataMethods,
    ModuleMethods,
):
    """
    ClientAPI is a medium-level wrapper around the HTTPAPI that provides many easy-to-use
    functions for accessing the client-server API.

    This class can be used directly, but generally you should use the higher-level wrappers that
    inherit from this class, such as :class:`mautrix.client.Client`
    or :class:`mautrix.appservice.IntentAPI`.

    Examples:
        >>> from mautrix.client import ClientAPI
        >>> client = ClientAPI("@user:matrix.org", base_url="https://matrix-client.matrix-org",
                               token="syt_123_456")
        >>> await client.whoami()
        WhoamiResponse(user_id="@user:matrix.org", device_id="DEV123")
        >>> await client.get_joined_rooms()
        ["!roomid:matrix.org"]
    """



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

import warnings

from mautrix.types import IdentityKey, SessionID

from .base import MatrixError


class CryptoError(MatrixError):
    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.message = message


class EncryptionError(CryptoError):
    pass


class SessionShareError(CryptoError):
    pass


class DecryptionError(CryptoError):
    @property
    def human_message(self) -> str:
        return "the bridge failed to decrypt the message"


class MatchingSessionDecryptionError(DecryptionError):
    pass


class GroupSessionWithheldError(DecryptionError):
    def __init__(self, session_id: SessionID, withheld_code: str) -> None:
        super().__init__(f"Session ID {session_id} was withheld ({withheld_code})")
        self.withheld_code = withheld_code


class SessionNotFound(DecryptionError):
    def __init__(self, session_id: SessionID, sender_key: IdentityKey | None = None) -> None:
        super().__init__(
            f"Failed to decrypt megolm event: no session with given ID {session_id} found"
        )
        self.session_id = session_id
        self._sender_key = sender_key

    @property
    def human_message(self) -> str:
        return "the bridge hasn't received the decryption keys"

    @property
    def sender_key(self) -> IdentityKey | None:
        """
        .. deprecated:: 0.17.0
            Matrix v1.3 deprecated the device_id and sender_key fields in megolm events.
        """
        warnings.warn(
            "The sender_key field in Megolm events was deprecated in Matrix 1.3",
            DeprecationWarning,
        )
        return self._sender_key


class DuplicateMessageIndex(DecryptionError):
    def __init__(self) -> None:
        super().__init__("Duplicate message index")


class VerificationError(DecryptionError):
    def __init__(self) -> None:
        super().__init__("Device keys in session and cached device info do not match")


class DecryptedPayloadError(DecryptionError):
    pass


class MismatchingRoomError(DecryptionError):
    def __init__(self) -> None:
        super().__init__("Encrypted megolm event is not intended for this room")


class DeviceValidationError(EncryptionError):
    pass



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from typing import List, Sequence, Union

from .formatted_string import EntityType, FormattedString


class MarkdownString(FormattedString):
    text: str

    def __init__(self, text: str = "") -> None:
        self.text = text

    def __str__(self) -> str:
        return self.text

    def append(self, *args: Union[str, FormattedString]) -> MarkdownString:
        self.text += "".join(str(arg) for arg in args)
        return self

    def prepend(self, *args: Union[str, FormattedString]) -> MarkdownString:
        self.text = "".join(str(arg) for arg in args + (self.text,))
        return self

    def format(self, entity_type: EntityType, **kwargs) -> MarkdownString:
        if entity_type == EntityType.BOLD:
            self.text = f"**{self.text}**"
        elif entity_type == EntityType.ITALIC:
            self.text = f"_{self.text}_"
        elif entity_type == EntityType.STRIKETHROUGH:
            self.text = f"~~{self.text}~~"
        elif entity_type == EntityType.SPOILER:
            reason = kwargs.get("reason", "")
            if reason:
                self.text = f"{reason}|{self.text}"
            self.text = f"||{self.text}||"
        elif entity_type == EntityType.URL:
            if kwargs["url"] != self.text:
                self.text = f"[{self.text}]({kwargs['url']})"
        elif entity_type == EntityType.PREFORMATTED:
            self.text = f"```{kwargs['language']}\n{self.text}\n```"
        elif entity_type == EntityType.INLINE_CODE:
            self.text = f"`{self.text}`"
        elif entity_type == EntityType.BLOCKQUOTE:
            children = self.trim().split("\n")
            children = [child.prepend("> ") for child in children]
            self.text = self.join(children, "\n").text
        elif entity_type == EntityType.HEADER:
            prefix = "#" * kwargs["size"]
            self.text = f"{prefix} {self.text}"

        return self

    def trim(self) -> MarkdownString:
        self.text = self.text.strip()
        return self

    def split(self, separator, max_items: int = -1) -> List[MarkdownString]:
        return [MarkdownString(text) for text in self.text.split(separator, max_items)]

    @classmethod
    def join(
        cls, items: Sequence[Union[str, FormattedString]], separator: str = " "
    ) -> MarkdownString:
        return cls(separator.join(str(item) for item in items))



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import List, NewType, Optional, Union

from attr import dataclass

from .primitive import JSON, DeviceID, UserID
from .util import ExtensibleEnum, Obj, SerializableAttrs, deserializer, field


class LoginType(ExtensibleEnum):
    """
    A login type, as specified in the `POST /login endpoint`_

    .. _POST /login endpoint:
        https://spec.matrix.org/v1.2/client-server-api/#post_matrixclientv3login
    """

    PASSWORD: "LoginType" = "m.login.password"
    TOKEN: "LoginType" = "m.login.token"
    SSO: "LoginType" = "m.login.sso"
    APPSERVICE: "LoginType" = "m.login.application_service"

    UNSTABLE_JWT: "LoginType" = "org.matrix.login.jwt"


@dataclass
class LoginFlow(SerializableAttrs):
    """
    A login flow, as specified in the `GET /login endpoint`_

    .. _GET /login endpoint:
        https://spec.matrix.org/v1.2/client-server-api/#get_matrixclientv3login
    """

    type: LoginType


@dataclass
class LoginFlowList(SerializableAttrs):
    flows: List[LoginFlow]

    def get_first_of_type(self, *types: LoginType) -> Optional[LoginFlow]:
        for flow in self.flows:
            if flow.type in types:
                return flow
        return None

    def supports_type(self, *types: LoginType) -> bool:
        return self.get_first_of_type(*types) is not None


class UserIdentifierType(ExtensibleEnum):
    """
    A user identifier type, as specified in the `Identifier types`_ section of the login spec.

    .. _Identifier types:
        https://spec.matrix.org/v1.2/client-server-api/#identifier-types
    """

    MATRIX_USER: "UserIdentifierType" = "m.id.user"
    THIRD_PARTY: "UserIdentifierType" = "m.id.thirdparty"
    PHONE: "UserIdentifierType" = "m.id.phone"


@dataclass
class MatrixUserIdentifier(SerializableAttrs):
    """
    A client can identify a user using their Matrix ID. This can either be the fully qualified
    Matrix user ID, or just the localpart of the user ID.
    """

    user: str
    """The Matrix user ID or localpart"""

    type: UserIdentifierType = UserIdentifierType.MATRIX_USER


@dataclass
class ThirdPartyIdentifier(SerializableAttrs):
    """
    A client can identify a user using a 3PID associated with the user's account on the homeserver,
    where the 3PID was previously associated using the `/account/3pid`_ API. See the `3PID Types`_
    Appendix for a list of Third-party ID media.

    .. _/account/3pid:
        https://spec.matrix.org/v1.2/client-server-api/#post_matrixclientv3account3pid
    .. _3PID Types:
        https://spec.matrix.org/v1.2/appendices/#3pid-types
    """

    medium: str
    address: str
    type: UserIdentifierType = UserIdentifierType.THIRD_PARTY


@dataclass
class PhoneIdentifier(SerializableAttrs):
    """
    A client can identify a user using a phone number associated with the user's account, where the
    phone number was previously associated using the `/account/3pid`_ API. The phone number can be
    passed in as entered by the user; the homeserver will be responsible for canonicalising it.
    If the client wishes to canonicalise the phone number, then it can use the ``m.id.thirdparty``
    identifier type with a ``medium`` of ``msisdn`` instead.

    .. _/account/3pid:
        https://spec.matrix.org/v1.2/client-server-api/#post_matrixclientv3account3pid
    """

    country: str
    phone: str
    type: UserIdentifierType = UserIdentifierType.PHONE


UserIdentifier = NewType(
    "UserIdentifier", Union[MatrixUserIdentifier, ThirdPartyIdentifier, PhoneIdentifier]
)


@deserializer(UserIdentifier)
def deserialize_user_identifier(data: JSON) -> Union[UserIdentifier, Obj]:
    try:
        identifier_type = UserIdentifierType.deserialize(data["type"])
    except KeyError:
        return Obj(**data)
    if identifier_type == UserIdentifierType.MATRIX_USER:
        return MatrixUserIdentifier.deserialize(data)
    elif identifier_type == UserIdentifierType.THIRD_PARTY:
        return ThirdPartyIdentifier.deserialize(data)
    elif identifier_type == UserIdentifierType.PHONE:
        return PhoneIdentifier.deserialize(data)
    else:
        return Obj(**data)


setattr(UserIdentifier, "deserialize", deserialize_user_identifier)


@dataclass
class DiscoveryServer(SerializableAttrs):
    base_url: Optional[str] = None


@dataclass
class DiscoveryIntegrationServer(SerializableAttrs):
    ui_url: Optional[str] = None
    api_url: Optional[str] = None


@dataclass
class DiscoveryIntegrations(SerializableAttrs):
    managers: List[DiscoveryIntegrationServer] = field(factory=lambda: [])


@dataclass
class DiscoveryInformation(SerializableAttrs):
    """
    .well-known discovery information, as specified in the `GET /.well-known/matrix/client endpoint`_

    .. _GET /.well-known/matrix/client endpoint:
        https://spec.matrix.org/v1.2/client-server-api/#getwell-knownmatrixclient
    """

    homeserver: Optional[DiscoveryServer] = field(json="m.homeserver", factory=DiscoveryServer)
    identity_server: Optional[DiscoveryServer] = field(
        json="m.identity_server", factory=DiscoveryServer
    )
    integrations: Optional[DiscoveryServer] = field(
        json="m.integrations", factory=DiscoveryIntegrations
    )


@dataclass
class LoginResponse(SerializableAttrs):
    """
    The response for a login request, as specified in the `POST /login endpoint`_

    .. _POST /login endpoint:
        https://spec.matrix.org/v1.2/client-server-api/#post_matrixclientv3login
    """

    user_id: UserID
    device_id: DeviceID
    access_token: str
    well_known: DiscoveryInformation = field(factory=DiscoveryInformation)


@dataclass
class WhoamiResponse(SerializableAttrs):
    """
    The response for a whoami request, as specified in the `GET /account/whoami endpoint`_

    .. _GET /account/whoami endpoint:
        https://spec.matrix.org/v1.2/client-server-api/#get_matrixclientv3accountwhoami
    """

    user_id: UserID
    device_id: Optional[DeviceID] = None
    is_guest: bool = False



# Copyright (c) 2023 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

import logging

from mautrix.util.async_db import Connection, Scheme, UpgradeTable

upgrade_table = UpgradeTable(
    version_table_name="crypto_version",
    database_name="crypto store",
    log=logging.getLogger("mau.crypto.db.upgrade"),
)


@upgrade_table.register(description="Latest revision", upgrades_to=10)
async def upgrade_blank_to_latest(conn: Connection) -> None:
    await conn.execute(
        """CREATE TABLE IF NOT EXISTS crypto_account (
            account_id TEXT    PRIMARY KEY,
            device_id  TEXT    NOT NULL,
            shared     BOOLEAN NOT NULL,
            sync_token TEXT    NOT NULL,
            account    bytea   NOT NULL
        )"""
    )
    await conn.execute(
        """CREATE TABLE IF NOT EXISTS crypto_message_index (
            sender_key CHAR(43),
            session_id CHAR(43),
            "index"    INTEGER,
            event_id   TEXT   NOT NULL,
            timestamp  BIGINT NOT NULL,
            PRIMARY KEY (sender_key, session_id, "index")
        )"""
    )
    await conn.execute(
        """CREATE TABLE IF NOT EXISTS crypto_tracked_user (
            user_id TEXT PRIMARY KEY
        )"""
    )
    await conn.execute(
        """CREATE TABLE IF NOT EXISTS crypto_device (
            user_id      TEXT,
            device_id    TEXT,
            identity_key CHAR(43) NOT NULL,
            signing_key  CHAR(43) NOT NULL,
            trust        SMALLINT NOT NULL,
            deleted      BOOLEAN  NOT NULL,
            name         TEXT     NOT NULL,
            PRIMARY KEY (user_id, device_id)
        )"""
    )
    await conn.execute(
        """CREATE TABLE IF NOT EXISTS crypto_olm_session (
            account_id     TEXT,
            session_id     CHAR(43),
            sender_key     CHAR(43)  NOT NULL,
            session        bytea     NOT NULL,
            created_at     timestamp NOT NULL,
            last_decrypted timestamp NOT NULL,
            last_encrypted timestamp NOT NULL,
            PRIMARY KEY (account_id, session_id)
        )"""
    )
    await conn.execute(
        """CREATE TABLE IF NOT EXISTS crypto_megolm_inbound_session (
            account_id        TEXT,
            session_id        CHAR(43),
            sender_key        CHAR(43)  NOT NULL,
            signing_key       CHAR(43),
            room_id           TEXT      NOT NULL,
            session           bytea,
            forwarding_chains TEXT,
            withheld_code     TEXT,
            withheld_reason   TEXT,
            ratchet_safety    jsonb,
            received_at       timestamp,
            max_age           BIGINT,
            max_messages      INTEGER,
            is_scheduled      BOOLEAN NOT NULL DEFAULT false,
            PRIMARY KEY (account_id, session_id)
        )"""
    )
    await conn.execute(
        """CREATE TABLE IF NOT EXISTS crypto_megolm_outbound_session (
            account_id    TEXT,
            room_id       TEXT,
            session_id    CHAR(43)  NOT NULL UNIQUE,
            session       bytea     NOT NULL,
            shared        BOOLEAN   NOT NULL,
            max_messages  INTEGER   NOT NULL,
            message_count INTEGER   NOT NULL,
            max_age       BIGINT    NOT NULL,
            created_at    timestamp NOT NULL,
            last_used     timestamp NOT NULL,
            PRIMARY KEY (account_id, room_id)
        )"""
    )
    await conn.execute(
        """CREATE TABLE crypto_cross_signing_keys (
            user_id TEXT,
            usage   TEXT,
            key     CHAR(43) NOT NULL,

            first_seen_key CHAR(43) NOT NULL,

            PRIMARY KEY (user_id, usage)
        )"""
    )
    await conn.execute(
        """CREATE TABLE crypto_cross_signing_signatures (
            signed_user_id TEXT,
            signed_key     TEXT,
            signer_user_id TEXT,
            signer_key     TEXT,
            signature      CHAR(88) NOT NULL,
            PRIMARY KEY (signed_user_id, signed_key, signer_user_id, signer_key)
        )"""
    )


@upgrade_table.register(description="Add account_id primary key column")
async def upgrade_v2(conn: Connection, scheme: Scheme) -> None:
    if scheme == Scheme.SQLITE:
        await conn.execute("DROP TABLE crypto_account")
        await conn.execute("DROP TABLE crypto_olm_session")
        await conn.execute("DROP TABLE crypto_megolm_inbound_session")
        await conn.execute("DROP TABLE crypto_megolm_outbound_session")
        await conn.execute(
            """CREATE TABLE crypto_account (
                account_id VARCHAR(255) PRIMARY KEY,
                device_id  VARCHAR(255) NOT NULL,
                shared     BOOLEAN      NOT NULL,
                sync_token TEXT         NOT NULL,
                account    bytea        NOT NULL
            )"""
        )
        await conn.execute(
            """CREATE TABLE crypto_olm_session (
                account_id   VARCHAR(255),
                session_id   CHAR(43),
                sender_key   CHAR(43)  NOT NULL,
                session      bytea     NOT NULL,
                created_at   timestamp NOT NULL,
                last_used    timestamp NOT NULL,
                PRIMARY KEY (account_id, session_id)
            )"""
        )
        await conn.execute(
            """CREATE TABLE crypto_megolm_inbound_session (
                account_id   VARCHAR(255),
                session_id   CHAR(43),
                sender_key   CHAR(43)     NOT NULL,
                signing_key  CHAR(43)     NOT NULL,
                room_id      VARCHAR(255) NOT NULL,
                session      bytea        NOT NULL,
                forwarding_chains TEXT    NOT NULL,
                PRIMARY KEY (account_id, session_id)
            )"""
        )
        await conn.execute(
            """CREATE TABLE crypto_megolm_outbound_session (
                account_id    VARCHAR(255),
                room_id       VARCHAR(255),
                session_id    CHAR(43)     NOT NULL UNIQUE,
                session       bytea        NOT NULL,
                shared        BOOLEAN      NOT NULL,
                max_messages  INTEGER      NOT NULL,
                message_count INTEGER      NOT NULL,
                max_age       BIGINT       NOT NULL,
                created_at    timestamp    NOT NULL,
                last_used     timestamp    NOT NULL,
                PRIMARY KEY (account_id, room_id)
            )"""
        )
    else:

        async def add_account_id_column(table: str, pkey_columns: list[str]) -> None:
            await conn.execute(f"ALTER TABLE {table} ADD COLUMN account_id VARCHAR(255)")
            await conn.execute(f"UPDATE {table} SET account_id=''")
            await conn.execute(f"ALTER TABLE {table} ALTER COLUMN account_id SET NOT NULL")
            await conn.execute(f"ALTER TABLE {table} DROP CONSTRAINT {table}_pkey")
            pkey_columns.append("account_id")
            pkey_columns_str = ", ".join(f'"{col}"' for col in pkey_columns)
            await conn.execute(
                f"ALTER TABLE {table} ADD CONSTRAINT {table}_pkey "
                f"PRIMARY KEY ({pkey_columns_str})"
            )

        await add_account_id_column("crypto_account", [])
        await add_account_id_column("crypto_olm_session", ["session_id"])
        await add_account_id_column("crypto_megolm_inbound_session", ["session_id"])
        await add_account_id_column("crypto_megolm_outbound_session", ["room_id"])


@upgrade_table.register(description="Stop using size-limited string fields")
async def upgrade_v3(conn: Connection, scheme: Scheme) -> None:
    if scheme == Scheme.SQLITE:
        return
    await conn.execute("ALTER TABLE crypto_account ALTER COLUMN account_id TYPE TEXT")
    await conn.execute("ALTER TABLE crypto_account ALTER COLUMN device_id TYPE TEXT")
    await conn.execute("ALTER TABLE crypto_message_index ALTER COLUMN event_id TYPE TEXT")
    await conn.execute("ALTER TABLE crypto_tracked_user ALTER COLUMN user_id TYPE TEXT")
    await conn.execute("ALTER TABLE crypto_device ALTER COLUMN user_id TYPE TEXT")
    await conn.execute("ALTER TABLE crypto_device ALTER COLUMN device_id TYPE TEXT")
    await conn.execute("ALTER TABLE crypto_device ALTER COLUMN name TYPE TEXT")
    await conn.execute("ALTER TABLE crypto_olm_session ALTER COLUMN account_id TYPE TEXT")
    await conn.execute(
        "ALTER TABLE crypto_megolm_inbound_session ALTER COLUMN account_id TYPE TEXT"
    )
    await conn.execute("ALTER TABLE crypto_megolm_inbound_session ALTER COLUMN room_id TYPE TEXT")
    await conn.execute(
        "ALTER TABLE crypto_megolm_outbound_session ALTER COLUMN account_id TYPE TEXT"
    )
    await conn.execute("ALTER TABLE crypto_megolm_outbound_session ALTER COLUMN room_id TYPE TEXT")


@upgrade_table.register(description="Split last_used into last_encrypted and last_decrypted")
async def upgrade_v4(conn: Connection, scheme: Scheme) -> None:
    await conn.execute("ALTER TABLE crypto_olm_session RENAME COLUMN last_used TO last_decrypted")
    await conn.execute("ALTER TABLE crypto_olm_session ADD COLUMN last_encrypted timestamp")
    await conn.execute("UPDATE crypto_olm_session SET last_encrypted=last_decrypted")
    if scheme == Scheme.POSTGRES:
        # This is too hard to do on sqlite, so let's just do it on postgres
        await conn.execute(
            "ALTER TABLE crypto_olm_session ALTER COLUMN last_encrypted SET NOT NULL"
        )


@upgrade_table.register(description="Add cross-signing key and signature caches")
async def upgrade_v5(conn: Connection) -> None:
    await conn.execute(
        """CREATE TABLE crypto_cross_signing_keys (
            user_id TEXT,
            usage   TEXT,
            key     CHAR(43),
            first_seen_key CHAR(43),
            PRIMARY KEY (user_id, usage)
        )"""
    )
    await conn.execute(
        """CREATE TABLE crypto_cross_signing_signatures (
            signed_user_id TEXT,
            signed_key     TEXT,
            signer_user_id TEXT,
            signer_key     TEXT,
            signature      TEXT,
            PRIMARY KEY (signed_user_id, signed_key, signer_user_id, signer_key)
        )"""
    )


@upgrade_table.register(description="Update trust state values")
async def upgrade_v6(conn: Connection) -> None:
    await conn.execute("UPDATE crypto_device SET trust=300 WHERE trust=1")  # verified
    await conn.execute("UPDATE crypto_device SET trust=-100 WHERE trust=2")  # blacklisted
    await conn.execute("UPDATE crypto_device SET trust=0 WHERE trust=3")  # ignored -> unset


@upgrade_table.register(
    description="Synchronize schema with mautrix-go", upgrades_to=9, transaction=False
)
async def upgrade_v9(conn: Connection, scheme: Scheme) -> None:
    if scheme == Scheme.POSTGRES:
        async with conn.transaction():
            await upgrade_v9_postgres(conn)
    else:
        await upgrade_v9_sqlite(conn)


# These two are never used because the previous one jumps from 6 to 9.
@upgrade_table.register
async def upgrade_noop_7_to_8(_: Connection) -> None:
    pass


@upgrade_table.register
async def upgrade_noop_8_to_9(_: Connection) -> None:
    pass


async def upgrade_v9_postgres(conn: Connection) -> None:
    await conn.execute("UPDATE crypto_account SET device_id='' WHERE device_id IS NULL")
    await conn.execute("ALTER TABLE crypto_account ALTER COLUMN device_id SET NOT NULL")

    await conn.execute(
        "ALTER TABLE crypto_megolm_inbound_session ALTER COLUMN signing_key DROP NOT NULL"
    )
    await conn.execute(
        "ALTER TABLE crypto_megolm_inbound_session ALTER COLUMN session DROP NOT NULL"
    )
    await conn.execute(
        "ALTER TABLE crypto_megolm_inbound_session ALTER COLUMN forwarding_chains DROP NOT NULL"
    )
    await conn.execute("ALTER TABLE crypto_megolm_inbound_session ADD COLUMN withheld_code TEXT")
    await conn.execute("ALTER TABLE crypto_megolm_inbound_session ADD COLUMN withheld_reason TEXT")

    await conn.execute("DELETE FROM crypto_cross_signing_keys WHERE key IS NULL")
    await conn.execute(
        "UPDATE crypto_cross_signing_keys SET first_seen_key=key WHERE first_seen_key IS NULL"
    )
    await conn.execute("ALTER TABLE crypto_cross_signing_keys ALTER COLUMN key SET NOT NULL")
    await conn.execute(
        "ALTER TABLE crypto_cross_signing_keys ALTER COLUMN first_seen_key SET NOT NULL"
    )

    await conn.execute("DELETE FROM crypto_cross_signing_signatures WHERE signature IS NULL")
    await conn.execute(
        "ALTER TABLE crypto_cross_signing_signatures ALTER COLUMN signature SET NOT NULL"
    )

    await conn.execute(
        "ALTER TABLE crypto_megolm_outbound_session ALTER COLUMN max_age TYPE BIGINT "
        "USING (EXTRACT(EPOCH from max_age)*1000)::bigint"
    )


async def upgrade_v9_sqlite(conn: Connection) -> None:
    await conn.execute("PRAGMA foreign_keys = OFF")
    async with conn.transaction():
        await conn.execute(
            """CREATE TABLE new_crypto_account (
                account_id TEXT    PRIMARY KEY,
                device_id  TEXT    NOT NULL,
                shared     BOOLEAN NOT NULL,
                sync_token TEXT    NOT NULL,
                account    bytea   NOT NULL
            )"""
        )
        await conn.execute(
            """
            INSERT INTO new_crypto_account (account_id, device_id, shared, sync_token, account)
            SELECT account_id, COALESCE(device_id, ''), shared, sync_token, account
            FROM crypto_account
            """
        )
        await conn.execute("DROP TABLE crypto_account")
        await conn.execute("ALTER TABLE new_crypto_account RENAME TO crypto_account")

        await conn.execute(
            """CREATE TABLE new_crypto_megolm_inbound_session (
                account_id        TEXT,
                session_id        CHAR(43),
                sender_key        CHAR(43)  NOT NULL,
                signing_key       CHAR(43),
                room_id           TEXT      NOT NULL,
                session           bytea,
                forwarding_chains TEXT,
                withheld_code     TEXT,
                withheld_reason   TEXT,
                PRIMARY KEY (account_id, session_id)
            )"""
        )
        await conn.execute(
            """
            INSERT INTO new_crypto_megolm_inbound_session (
                account_id, session_id, sender_key, signing_key, room_id, session,
                forwarding_chains
            )
            SELECT account_id, session_id, sender_key, signing_key, room_id, session,
                   forwarding_chains
            FROM crypto_megolm_inbound_session
            """
        )
        await conn.execute("DROP TABLE crypto_megolm_inbound_session")
        await conn.execute(
            "ALTER TABLE new_crypto_megolm_inbound_session RENAME TO crypto_megolm_inbound_session"
        )

        await conn.execute("UPDATE crypto_megolm_outbound_session SET max_age=max_age*1000")

        await conn.execute(
            """CREATE TABLE new_crypto_cross_signing_keys (
                user_id TEXT,
                usage   TEXT,
                key     CHAR(43) NOT NULL,

                first_seen_key CHAR(43) NOT NULL,

                PRIMARY KEY (user_id, usage)
            )"""
        )
        await conn.execute(
            """
            INSERT INTO new_crypto_cross_signing_keys (user_id, usage, key, first_seen_key)
            SELECT user_id, usage, key, COALESCE(first_seen_key, key)
            FROM crypto_cross_signing_keys
            WHERE key IS NOT NULL
            """
        )
        await conn.execute("DROP TABLE crypto_cross_signing_keys")
        await conn.execute(
            "ALTER TABLE new_crypto_cross_signing_keys RENAME TO crypto_cross_signing_keys"
        )

        await conn.execute(
            """CREATE TABLE new_crypto_cross_signing_signatures (
                signed_user_id TEXT,
                signed_key     TEXT,
                signer_user_id TEXT,
                signer_key     TEXT,
                signature      CHAR(88) NOT NULL,
                PRIMARY KEY (signed_user_id, signed_key, signer_user_id, signer_key)
            )"""
        )
        await conn.execute(
            """
            INSERT INTO new_crypto_cross_signing_signatures (
                signed_user_id, signed_key, signer_user_id, signer_key, signature
            )
            SELECT signed_user_id, signed_key, signer_user_id, signer_key, signature
            FROM crypto_cross_signing_signatures
            WHERE signature IS NOT NULL
            """
        )
        await conn.execute("DROP TABLE crypto_cross_signing_signatures")
        await conn.execute(
            "ALTER TABLE new_crypto_cross_signing_signatures "
            "RENAME TO crypto_cross_signing_signatures"
        )

        await conn.execute("PRAGMA foreign_key_check")
    await conn.execute("PRAGMA foreign_keys = ON")


@upgrade_table.register(
    description="Add metadata for detecting when megolm sessions are safe to delete"
)
async def upgrade_v10(conn: Connection) -> None:
    await conn.execute("ALTER TABLE crypto_megolm_inbound_session ADD COLUMN ratchet_safety jsonb")
    await conn.execute(
        "ALTER TABLE crypto_megolm_inbound_session ADD COLUMN received_at timestamp"
    )
    await conn.execute("ALTER TABLE crypto_megolm_inbound_session ADD COLUMN max_age BIGINT")
    await conn.execute("ALTER TABLE crypto_megolm_inbound_session ADD COLUMN max_messages INTEGER")
    await conn.execute(
        "ALTER TABLE crypto_megolm_inbound_session "
        "ADD COLUMN is_scheduled BOOLEAN NOT NULL DEFAULT false"
    )



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from .account_data import AccountDataMethods
from .crypto import CryptoMethods
from .media_repository import MediaRepositoryMethods
from .misc import MiscModuleMethods
from .push_rules import PushRuleMethods
from .room_tag import RoomTaggingMethods


class ModuleMethods(
    MediaRepositoryMethods,
    CryptoMethods,
    AccountDataMethods,
    MiscModuleMethods,
    RoomTaggingMethods,
    PushRuleMethods,
):
    """
    Methods in section 13 Modules of the spec.

    See also: `API reference <https://matrix.org/docs/spec/client_server/r0.6.1.html#modules>`__
    """

    # TODO: subregions 15, 21, 26, 27, others?



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from typing import Optional

from attr import dataclass

from .primitive import ContentURI
from .util import SerializableAttrs, field


@dataclass
class MediaRepoConfig(SerializableAttrs):
    """
    Matrix media repo config. See `GET /_matrix/media/v3/config`_.

    .. _GET /_matrix/media/v3/config:
        https://spec.matrix.org/v1.2/client-server-api/#get_matrixmediav3config
    """

    upload_size: int = field(default=50 * 1024 * 1024, json="m.upload.size")


@dataclass
class OpenGraphImage(SerializableAttrs):
    url: ContentURI = field(default=None, json="og:image")
    mimetype: str = field(default=None, json="og:image:type")
    height: int = field(default=None, json="og:image:width")
    width: int = field(default=None, json="og:image:height")
    size: int = field(default=None, json="matrix:image:size")


@dataclass
class OpenGraphVideo(SerializableAttrs):
    url: ContentURI = field(default=None, json="og:video")
    mimetype: str = field(default=None, json="og:video:type")
    height: int = field(default=None, json="og:video:width")
    width: int = field(default=None, json="og:video:height")
    size: int = field(default=None, json="matrix:video:size")


@dataclass
class OpenGraphAudio(SerializableAttrs):
    url: ContentURI = field(default=None, json="og:audio")
    mimetype: str = field(default=None, json="og:audio:type")


@dataclass
class MXOpenGraph(SerializableAttrs):
    """
    Matrix URL preview response. See `GET /_matrix/media/v3/preview_url`_.

    .. _GET /_matrix/media/v3/preview_url:
        https://spec.matrix.org/v1.2/client-server-api/#get_matrixmediav3preview_url
    """

    title: str = field(default=None, json="og:title")
    description: str = field(default=None, json="og:description")
    image: OpenGraphImage = field(default=None, flatten=True)
    video: OpenGraphVideo = field(default=None, flatten=True)
    audio: OpenGraphAudio = field(default=None, flatten=True)


@dataclass
class MediaCreateResponse(SerializableAttrs):
    """
    Matrix media create response including MSC3870
    """

    content_uri: ContentURI
    unused_expired_at: Optional[int] = None
    unstable_upload_url: Optional[str] = field(default=None, json="com.beeper.msc3870.upload_url")



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

import asyncio
import logging

from mautrix import __optional_imports__
from mautrix.errors import DecryptionError, EncryptionError, MNotFound
from mautrix.types import (
    EncryptedEvent,
    EncryptedMegolmEventContent,
    EventContent,
    EventID,
    EventType,
    RoomID,
)
from mautrix.util.logging import TraceLogger

from . import client, dispatcher, store_updater

if __optional_imports__:
    from .. import crypto as crypt


class EncryptingAPI(store_updater.StoreUpdatingAPI):
    """
    EncryptingAPI is a wrapper around StoreUpdatingAPI that automatically encrypts messages.

    For automatic decryption, see :class:`DecryptionDispatcher`.
    """

    _crypto: crypt.OlmMachine | None
    encryption_blacklist: set[EventType] = {EventType.REACTION}
    """A set of event types which shouldn't be encrypted even in encrypted rooms."""
    crypto_log: TraceLogger = logging.getLogger("mau.client.crypto")
    """The logger to use for crypto-related things."""
    _share_session_events: dict[RoomID, asyncio.Event]

    def __init__(self, *args, crypto_log: TraceLogger | None = None, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        if crypto_log:
            self.crypto_log = crypto_log
        self._crypto = None
        self._share_session_events = {}

    @property
    def crypto(self) -> crypt.OlmMachine | None:
        """The :class:`crypto.OlmMachine` to use for e2ee stuff."""
        return self._crypto

    @crypto.setter
    def crypto(self, crypto: crypt.OlmMachine) -> None:
        """
        Args:
            crypto: The olm machine to use for crypto

        Raises:
            ValueError: if :attr:`state_store` is not set.
        """
        if not self.state_store:
            raise ValueError("State store must be set to use encryption")
        self._crypto = crypto

    @property
    def crypto_enabled(self) -> bool:
        """``True`` if both the olm machine and state store are set properly."""
        return bool(self.crypto) and bool(self.state_store)

    async def encrypt(
        self, room_id: RoomID, event_type: EventType, content: EventContent
    ) -> EncryptedMegolmEventContent:
        """
        Encrypt a message for the given room. Automatically creates and shares a group session
        if necessary.

        Args:
            room_id: The room to encrypt the event to.
            event_type: The type of event.
            content: The content of the event.

        Returns:
            The content of the encrypted event.
        """
        try:
            return await self.crypto.encrypt_megolm_event(room_id, event_type, content)
        except EncryptionError:
            self.crypto_log.debug("Got EncryptionError, sharing group session and trying again")
            await self.share_group_session(room_id)
            self.crypto_log.trace(
                f"Shared group session, now trying to encrypt in {room_id} again"
            )
            return await self.crypto.encrypt_megolm_event(room_id, event_type, content)

    async def _share_session_lock(self, room_id: RoomID) -> bool:
        try:
            event = self._share_session_events[room_id]
        except KeyError:
            self._share_session_events[room_id] = asyncio.Event()
            return True
        else:
            await event.wait()
            return False

    async def share_group_session(self, room_id: RoomID) -> None:
        """
        Create and share a Megolm session for the given room.

        Args:
            room_id: The room to share the session for.
        """
        if not await self._share_session_lock(room_id):
            self.log.silly("Group session was already being shared, so didn't share new one")
            return
        try:
            if not await self.state_store.has_full_member_list(room_id):
                self.crypto_log.trace(
                    f"Don't have full member list for {room_id}, fetching from server"
                )
                members = list((await self.get_joined_members(room_id)).keys())
            else:
                self.crypto_log.trace(f"Fetching member list for {room_id} from state store")
                members = await self.state_store.get_members(room_id)
            await self.crypto.share_group_session(room_id, members)
        finally:
            self._share_session_events.pop(room_id).set()

    async def send_message_event(
        self,
        room_id: RoomID,
        event_type: EventType,
        content: EventContent,
        disable_encryption: bool = False,
        **kwargs,
    ) -> EventID:
        """
        A wrapper around :meth:`ClientAPI.send_message_event` that encrypts messages if the target
        room is encrypted.

        Args:
            room_id: The room to send the message to.
            event_type: The unencrypted event type.
            content: The unencrypted event content.
            disable_encryption: Set to ``True`` if you want to force-send an unencrypted message.
            **kwargs: Additional parameters to pass to :meth:`ClientAPI.send_message_event`.

        Returns:
            The ID of the event that was sent.
        """
        if self.crypto and event_type not in self.encryption_blacklist and not disable_encryption:
            is_encrypted = await self.state_store.is_encrypted(room_id)
            if is_encrypted is None:
                try:
                    await self.get_state_event(room_id, EventType.ROOM_ENCRYPTION)
                    is_encrypted = True
                except MNotFound:
                    is_encrypted = False
            if is_encrypted:
                content = await self.encrypt(room_id, event_type, content)
                event_type = EventType.ROOM_ENCRYPTED
        return await super().send_message_event(room_id, event_type, content, **kwargs)


class DecryptionDispatcher(dispatcher.SimpleDispatcher):
    """
    DecryptionDispatcher is a dispatcher that can be used with a :class:`client.Syncer`
    to automatically decrypt events and dispatch the unencrypted versions for event handlers.

    The easiest way to use this is with :class:`client.Client`, which automatically registers
    this dispatcher when :attr:`EncryptingAPI.crypto` is set.
    """

    event_type = EventType.ROOM_ENCRYPTED
    client: client.Client

    async def handle(self, evt: EncryptedEvent) -> None:
        try:
            self.client.crypto_log.trace(f"Decrypting {evt.event_id} in {evt.room_id}...")
            decrypted = await self.client.crypto.decrypt_megolm_event(evt)
        except DecryptionError as e:
            self.client.crypto_log.warning(f"Failed to decrypt {evt.event_id}: {e}")
            return
        self.client.crypto_log.trace(f"Decrypted {evt.event_id}: {decrypted}")
        self.client.dispatch_event(decrypted, evt.source)



from .abstract import StateStore
from .file import FileStateStore
from .memory import MemoryStateStore
from .sync import MemorySyncStore, SyncStore

__all__ = [
    "StateStore",
    "FileStateStore",
    "MemoryStateStore",
    "MemorySyncStore",
    "SyncStore",
    "asyncpg",
]



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

import asyncio

from mautrix.errors import MForbidden, MNotFound
from mautrix.types import (
    JSON,
    EventID,
    EventType,
    Member,
    Membership,
    MemberStateEventContent,
    RoomAlias,
    RoomID,
    StateEvent,
    StateEventContent,
    SyncToken,
    UserID,
)

from .api import ClientAPI
from .state_store import StateStore


class StoreUpdatingAPI(ClientAPI):
    """
    StoreUpdatingAPI is a wrapper around the medium-level ClientAPI that optionally updates
    a client state store with outgoing state events (after they're successfully sent).
    """

    state_store: StateStore | None

    def __init__(self, *args, state_store: StateStore | None = None, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.state_store = state_store

    async def join_room_by_id(
        self,
        room_id: RoomID,
        third_party_signed: JSON = None,
        extra_content: dict[str, JSON] | None = None,
    ) -> RoomID:
        room_id = await super().join_room_by_id(
            room_id, third_party_signed=third_party_signed, extra_content=extra_content
        )
        if room_id and not extra_content and self.state_store:
            await self.state_store.set_membership(room_id, self.mxid, Membership.JOIN)
        return room_id

    async def join_room(
        self,
        room_id_or_alias: RoomID | RoomAlias,
        servers: list[str] | None = None,
        third_party_signed: JSON = None,
        max_retries: int = 4,
    ) -> RoomID:
        room_id = await super().join_room(
            room_id_or_alias, servers, third_party_signed, max_retries
        )
        if room_id and self.state_store:
            await self.state_store.set_membership(room_id, self.mxid, Membership.JOIN)
        return room_id

    async def leave_room(
        self,
        room_id: RoomID,
        reason: str | None = None,
        extra_content: dict[str, JSON] | None = None,
        raise_not_in_room: bool = False,
    ) -> None:
        await super().leave_room(room_id, reason, extra_content, raise_not_in_room)
        if not extra_content and self.state_store:
            await self.state_store.set_membership(room_id, self.mxid, Membership.LEAVE)

    async def knock_room(
        self,
        room_id_or_alias: RoomID | RoomAlias,
        reason: str | None = None,
        servers: list[str] | None = None,
    ) -> RoomID:
        room_id = await super().knock_room(room_id_or_alias, reason, servers)
        if room_id and self.state_store:
            await self.state_store.set_membership(room_id, self.mxid, Membership.KNOCK)
        return room_id

    async def invite_user(
        self,
        room_id: RoomID,
        user_id: UserID,
        reason: str | None = None,
        extra_content: dict[str, JSON] | None = None,
    ) -> None:
        await super().invite_user(room_id, user_id, reason, extra_content=extra_content)
        if not extra_content and self.state_store:
            await self.state_store.set_membership(room_id, user_id, Membership.INVITE)

    async def kick_user(
        self,
        room_id: RoomID,
        user_id: UserID,
        reason: str = "",
        extra_content: dict[str, JSON] | None = None,
    ) -> None:
        await super().kick_user(room_id, user_id, reason=reason, extra_content=extra_content)
        if not extra_content and self.state_store:
            await self.state_store.set_membership(room_id, user_id, Membership.LEAVE)

    async def ban_user(
        self,
        room_id: RoomID,
        user_id: UserID,
        reason: str = "",
        extra_content: dict[str, JSON] | None = None,
    ) -> None:
        await super().ban_user(room_id, user_id, reason=reason, extra_content=extra_content)
        if not extra_content and self.state_store:
            await self.state_store.set_membership(room_id, user_id, Membership.BAN)

    async def unban_user(
        self,
        room_id: RoomID,
        user_id: UserID,
        reason: str = "",
        extra_content: dict[str, JSON] | None = None,
    ) -> None:
        await super().unban_user(room_id, user_id, reason=reason, extra_content=extra_content)
        if self.state_store:
            await self.state_store.set_membership(room_id, user_id, Membership.LEAVE)

    async def get_state(self, room_id: RoomID) -> list[StateEvent]:
        state = await super().get_state(room_id)
        if self.state_store:
            update_members = self.state_store.set_members(
                room_id,
                {evt.state_key: evt.content for evt in state if evt.type == EventType.ROOM_MEMBER},
            )
            await asyncio.gather(
                update_members,
                *[
                    self.state_store.update_state(evt)
                    for evt in state
                    if evt.type != EventType.ROOM_MEMBER
                ],
            )
        return state

    async def create_room(self, *args, **kwargs) -> RoomID:
        room_id = await super().create_room(*args, **kwargs)
        if self.state_store:
            invitee_membership = Membership.INVITE
            if kwargs.get("beeper_auto_join_invites"):
                invitee_membership = Membership.JOIN
            for user_id in kwargs.get("invitees", []):
                await self.state_store.set_membership(room_id, user_id, invitee_membership)
            for evt in kwargs.get("initial_state", []):
                await self.state_store.update_state(
                    StateEvent(
                        type=EventType.find(evt["type"], t_class=EventType.Class.STATE),
                        room_id=room_id,
                        event_id=EventID("$fake-create-id"),
                        sender=self.mxid,
                        state_key=evt.get("state_key", ""),
                        timestamp=0,
                        content=evt["content"],
                    )
                )
        return room_id

    async def send_state_event(
        self,
        room_id: RoomID,
        event_type: EventType,
        content: StateEventContent | dict[str, JSON],
        state_key: str = "",
        **kwargs,
    ) -> EventID:
        event_id = await super().send_state_event(
            room_id, event_type, content, state_key, **kwargs
        )
        if self.state_store:
            fake_event = StateEvent(
                type=event_type,
                room_id=room_id,
                event_id=event_id,
                sender=self.mxid,
                state_key=state_key,
                timestamp=0,
                content=content,
            )
            await self.state_store.update_state(fake_event)
        return event_id

    async def get_state_event(
        self, room_id: RoomID, event_type: EventType, state_key: str = ""
    ) -> StateEventContent:
        event = await super().get_state_event(room_id, event_type, state_key)
        if self.state_store:
            fake_event = StateEvent(
                type=event_type,
                room_id=room_id,
                event_id=EventID(""),
                sender=UserID(""),
                state_key=state_key,
                timestamp=0,
                content=event,
            )
            await self.state_store.update_state(fake_event)
        return event

    async def get_joined_members(self, room_id: RoomID) -> dict[UserID, Member]:
        members = await super().get_joined_members(room_id)
        if self.state_store:
            await self.state_store.set_members(room_id, members, only_membership=Membership.JOIN)
        return members

    async def get_members(
        self,
        room_id: RoomID,
        at: SyncToken | None = None,
        membership: Membership | None = None,
        not_membership: Membership | None = None,
    ) -> list[StateEvent]:
        member_events = await super().get_members(room_id, at, membership, not_membership)
        if self.state_store and not_membership != Membership.JOIN:
            await self.state_store.set_members(
                room_id,
                {evt.state_key: evt.content for evt in member_events},
                only_membership=membership,
            )
        return member_events

    async def fill_member_event(
        self,
        room_id: RoomID,
        user_id: UserID,
        content: MemberStateEventContent,
    ) -> MemberStateEventContent | None:
        """
        Fill a membership event content that is going to be sent in :meth:`send_member_event`.

        This is used to set default fields like the displayname and avatar, which are usually set
        by the server in the sugar membership endpoints like /join and /invite, but are not set
        automatically when sending member events manually.

        This implementation in StoreUpdatingAPI will first try to call the default implementation
        (which calls :attr:`fill_member_event_callback`). If that doesn't return anything, this
        will try to get the profile from the current member event, and then fall back to fetching
        the global profile from the server.

        Args:
            room_id: The room where the member event is going to be sent.
            user_id: The user whose membership is changing.
            content: The new member event content.

        Returns:
            The filled member event content.
        """
        callback_content = await super().fill_member_event(room_id, user_id, content)
        if callback_content is not None:
            self.log.trace("Filled new member event for %s using callback", user_id)
            return callback_content

        if content.displayname is None and content.avatar_url is None:
            existing_member = await self.state_store.get_member(room_id, user_id)
            if existing_member is not None:
                self.log.trace(
                    "Found existing member event %s to fill new member event for %s",
                    existing_member,
                    user_id,
                )
                content.displayname = existing_member.displayname
                content.avatar_url = existing_member.avatar_url
                return content

            try:
                profile = await self.get_profile(user_id)
            except (MNotFound, MForbidden):
                profile = None
            if profile:
                self.log.trace(
                    "Fetched profile %s to fill new member event of %s", profile, user_id
                )
                content.displayname = profile.displayname
                content.avatar_url = profile.avatar_url
                return content
            else:
                self.log.trace("Didn't find profile info to fill new member event of %s", user_id)
        else:
            self.log.trace(
                "Member event for %s already contains displayname or avatar, not re-filling",
                user_id,
            )
        return None



# Copyright (c) 2022 Tulir Asokan
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
from __future__ import annotations

from mautrix import __optional_imports__
from mautrix.types import Event, EventType, StateEvent

from .encryption_manager import DecryptionDispatcher, EncryptingAPI
from .state_store import StateStore, SyncStore
from .syncer import Syncer

if __optional_imports__:
    from .. import crypto as crypt


class Client(EncryptingAPI, Syncer):
    """Client is a high-level wrapper around the client API."""

    def __init__(
        self,
        *args,
        sync_store: SyncStore | None = None,
        state_store: StateStore | None = None,
        **kwargs,
    ) -> None:
        EncryptingAPI.__init__(self, *args, state_store=state_store, **kwargs)
        Syncer.__init__(self, sync_store)
        self.add_event_handler(EventType.ALL, self._update_state)

    async def _update_state(self, evt: Event) -> None:
        if not isinstance(evt, StateEvent) or not self.state_store:
            return
        await self.state_store.update_state(evt)

    @EncryptingAPI.crypto.setter
    def crypto(self, crypto: crypt.OlmMachine | None) -> None:
        """
        Set the olm machine and enable the automatic event decryptor.

        Args:
            crypto: The olm machine to use for crypto

        Raises:
            ValueError: if :attr:`state_store` is not set.
        """
        if not self.state_store:
            raise ValueError("State store must be set to use encryption")
        self._crypto = crypto
        if self.crypto_enabled:
            self.add_dispatcher(DecryptionDispatcher)
        else:
            self.remove_dispatcher(DecryptionDispatcher)
